/****************************************************************************
*
*      C implementation of EPFL - EKV (Enz-Krummenacher-Vittoz)  MOS  MODEL
*
*      VERSION             2.6 Revision III
*      DATE                June 2nd, 1997
*      LAST MODIFICATION   (r15) October 31, 2000
*      AUTHOR              Matthias Bucher                      LEG-DE-EPFL
*
*****************************************************************************/

/*  modifications:

00/10/31  mb  (r15) corrected dl, dw 'scalem' handling
00/09/15  mb  (r14) (gr)  reinstated sif_sir_3 and dQ_i_factor
99/30/09  mb  commented previous charges derivatives
99/09/09  mb  (r13) correction of deltaVGprime for compatibility v2.63/v2.62
              corrected 'epsilon' in sqrt_VGstar

99/07/30  mb  (r13 beta) introduced single equation VP expression for smooth
              charge/transcapacitances from depletion to accumulation
              (REVISION: III)
99/07/30  mb  (r12) corrected dGAMMAprime_dVG, dVP_dVG (narrow channel);
              preventing PHI from being smaller than 0.2 at init
              and after temperature update
99/05/10  mb  (r12) completed parameters init for XQC, DL, DW, removed
              IBC, ibc (cd)
99/03/26  mb  (r11) corrected multiple series device handling (asp)
              introduced M (synonym fo NP) and N (synonym for NS) as
              device parameters
98/10/11  mb  (r10) taking absolute value of effective field in mobility
              model (jpm)
98/08/28  mb  (r9) corrected PHI initialization, corrected UO and GAMMA
              units (jb)(REVISION: II)
98/08/26  mb  (r8) corrected UCRIT initialization (jb)
              introduced XJ initialization (rg), revised GAMMA initialization
98/08/20  mb  (r7) corrected VDSAT expression (operating point)
98/08/08  mb  (r6) corrected cgbi expression for capacitances only model (jpm)
              introduced calculation of electrical parameters as a function
              of new optional parameters TOX, NSUB, VFB, UO, VMAX
              THETA is now 'optional' if E0 is not specified
              added EO, QO as synonyms for E0, Q0
              removed UPDATE handling
98/07/09  mb  (r5) updated thermal noise/time constant calculation
			  simplified clauses for vertical field mobility calculation
              updated operating point information
98/05/14  mb  (r4) corrected IBB initialization, added clause to prevent
              substrate current calculation in case IBA <= 0
98/03/17  mb  (r3) introduced Vth, Vdss indication for .op information;
              removed description of extrinsic elements, and all related
              variables in data structures
98/02/20  mb  (r2) corrected UPDATE initialization, RSC handling
98/02/05  mb  (r1) corrected log_Vc_Vt
97/09/14  mb  introduced new symmetrized version of the bias dependent
              GAMMAprime (REVISION: I)
97/09/03  mb  corrected QB derivatives calculation, unified UPDATE handling and
              rearranged dynamic model code, completed TAU computation
97/08/21  mb  corrected GAMMAa usage (sa), Leff, Weff usage in LETA_L, WETA_W (rs)
              corrected mobility reduction model derivatives
97/07/18  mb  introduced EKV v2.6 short distance matching
              introduced inv_UCRIT, sqrt_PHI model int. variables
              introduced XQC, EKVDYN parameters for dynamic model setup
97/06/02  mb  introduced EKV v2.6 new mobility reduction model
              introduced UPDATE parameter for the choice of mobility models
97/05/16  mb  corrected units in rsce
97/04/21  mb  completed Isub derivatives
97/04/14  mb  introduced reverse short channel effect model
              and parameters (Q0, LK)  (v2.5)
97/02/10  mb  replaced F_v by F_v_n interpolation function
96/11/28  mb  suppr. deltaV; ddeltaV_factor init in 'ekv_initmodel';
			  completed analytical derivatives of charges
96/10/28  mb  simplified Vip, irprime derivatives
96/10/16  mb  simplified sif, sif2 calculation (jcp)
96/10/10  mb  changed *= to /= in rseff, rdeff calculation (ft)
96/08/23  mb  introduced new charges calculation (EKV v2.4)
96/08/23  mb  corrected parameter initialization VTO, TCV
              introduced N dependence of junction current w. temp.
              introduced interpolation function option EKVINT
96/07/15  mb  corrected thermal noise expression (ft)
              corrected VDSS*2 factor for substrate current
96/07/12  mb  introduced VTO sign attribution (mg)
96/06/20  mb  excluded sqrt_if recalculation (jp)
              corrected substrate current derivative (eb)
96/06/19  mb  excluded AF temp update (pm, eb)
96/04/29  mb  avoided sqrt_ir to be recomputed (eb)
96/04/23  mb  Lc_IBB, IBA_IBB moved to ekv_initmodel (pm)
96/04/12  mb  corrections Vt_4 calculation, Leq computation
95/12/19  mb  introduced code updates and impact ionization, code version 2.3
95/10/13  mb  corrected expression for slope factor n
95/08/29  mb  introduced new interpolation function, code version 2.2
95/05/24  mb  introduced sqrt_F() function instead of F()
95/05/16  mb  corrected sign in dVPprime_factor. Comp. ok.
95/05/14  mb  corrected Geo = mosgeo + itrans  (sa)
95/05/10  mb  changed model version to 2.1
              introduced new formulation for channel length mod. etc
              interpolation function (tabulated) replaced by simplified i.f.
			  minor corrections of typing errors
95/04/19  mb  corrected cgbi, completed various notes
95/04/13  mb  introduced SCALEM in default parameters assignment
              NRD, NRS default assignment using LDIF
              introduced more 'geo' variables
95/04/06  mb  corrected gmibd/s to gmb2/1
95/02/24  mb  overall code corrections
95/01/17  mb  output conductance model update, introduced PBSW
95/01/05  mb  introduced NFS, XJ parameters
95/01/04  mb  introduced temp. parameters: TR1, TR2 for RS/RD, XTI for IS
              introduced emission coefficient N
94/12/14  mb  simplified and  corrected temp. update of model parameters
              introduced JSW and TT parameters
              rearranged formulation of diode capacitances and access
              resistances

*/

/*----------------------------------------------------------------------------
*        __________________________________________
*
*        PLEASE READ THESE INSTRUCTIONS CAREFULLY!!
*        __________________________________________
*
*
*        Use of the EPFL-EKV MOST Model Code for Simulation
*        --------------------------------------------------
*
*        The use of this code is subject to written agreement of the modelling
*        group at the Swiss Federal Institute of Technology, Lausanne (EPFL).
*
*        This implementation is based on the equation set described in
*        'The EPFL-EKV MOSFET Model Equations for Simulation, Version 2.6',
*        M. Bucher, C. Lallement, F. Theodoloz, C. Enz, F. Krummenacher,
*        EPFL-DE-LEG, June, 1997
*
*        Theoretical discussion of the model can  be found in
*        C. Enz, F. Krummenacher, E. Vittoz, 'An analytical MOS transistor
*        model valid in all regions of Operation and dedicated to low-voltage
*        and low-current applications', Journal on Analog Integrated Circuits
*        and Signal Processsing, Kluwer Academic Publishers, pp. 83-114,
*        July 1995
*
*        This model is part of the public domain. Written consent of the
*        Ecole Polytechnique Federale de Lausanne (EPFL) is necessary for
*        use of this code. Third parties wishing to use this code or parts of
*        it for implementation in simulators, do have to agree to use the model
*        under its name EKV or EPFL-EKV MOST model. Any release of such
*        implementation should be authorized by EPFL, especially if it does not
*        contain all model features. EPFL can in no way be considered as being
*        responsible for damage to material or persons due to improper use of this
*        code or parts of it, or errors in this code or in the documentation.
*        Please contact the modelling group at EPFL for code error reports,
*        model & code upgrades or changes. This helps to keep this code as
*        'clean' as possible.
*
*        This code is not a stand-alone version for simulation, and adaptations
*        to particular environments are necessary. This version of the code
*        should be considered with care, although the C code has been compiled.
*        No guarantees can be given regarding its functionality. This code is
*        intended to give an idea of what a part of the implementation of the
*        EKV model could look like as a help for further implementations. The
*        EPFL-EKV model itself is constantly being improved and changes to the
*        model or its coding and thus its documentation may be frequent.
*        Version numbers are used to differentiate various releases. Backward
*        compatibility with previous releases is an aim but cannot be guaranteed.
*
*        Since the model is intended to be used in circuit simulators which
*        can handle a large number of transistors for various simulation models,
*        a particular data structure is proposed which helps to differentiate
*        what belongs to the simulation model (in this case, the EKV model)
*        and what is relied to the devices and their particular geometries.
*        To optimize the overall code and take advantage of this data structure,
*        the code parts have been 'distributed' into several functions, which
*        need to be called respectively for necessary initializations, or any
*        bias or temperature change, according to the needs of the circuit
*        simulator.
*
*        Included is the code for the intrinsic MOS model:
*          - static and quasi-static model equations,
*          - analytical computation of derivatives
*            (transconductances and transcapacitances),
*          - temperature effects,
*          - noise model equations.
*
*        The code regarding the extrinsic part of the MOS transistor is not
*        included here. Implementation in simulators should provide similar
*        functionality as commonly used for other MOS models.
*
*        Please contact:
*        ---------------
*        Matthias Bucher
*        matthias.bucher@epfl.ch
*        phone: ++41 21 693 46 02, fax: ++41 21 693 36 40
*        Electronics Laboratories ELB
*        Swiss Federal Institute of Technology, Lausanne (EPFL)
*        CH-1015 Lausanne, Switzerland
*
----------------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>

#include "ekv_v26_code_r15.h"


/*------------------------------------------------------------------------------*/
/*  Overall constants  */

#define  epso         8.854214871e-12
#define  epsox        3.4531438e-11
#define  epssil       1.035943140e-10
#define  boltz        1.3806226e-23
#define  charge       1.6021918e-19
#define  konq         boltz / charge
#define  twoq_esi     2.0*charge*epssil         /*  !!  mb  98/08/08  */
#define  twopi        6.2831853072e0

#define  reftmp       300.15e0
#define  tnomref      273.15e0
#define  Egreftmp     21.55280926               /*  !!  mb  98/08/08  */

#define  one3rd       1.0/3.0     /*  !!  mb  97/06/02  */
#define  two3rds      2.0/3.0
#define  four15th     4.0/15.0                  /*  !!  mb  96/07/23  ekv v2.4  */
#define  twentyfive16th  25.0/16.0


#define  c_a          0.28            /*  !!  mb  97/04/14  ekv v2.5, 97/05/16  corrected units  */
#define  c_ee         1.936e-3        /*  !!  mb  97/04/14  ekv v2.5  */

#define  false        0
#define  true         1


// moje def
#define model_index   0                     // indeks modelu w tablicy modeli - na razie uzywam tylko 1 dlatego
                                            // zamiast zmiennej imodel bede podstawiac stala model_index
#define trans_index   0                     // indeks - numer tranzystora w tablicy tranzystorow - na razie uzywam tylko dla 1
                                            // dlatego zamiast zmiennej itrans bede podstawiac stala trans_index
#define yes           1
#define no            0
// koniec moje def


/*------------------------------------------------------------------------------*/
/*  The following aliases declarations give the equivalences between the Model
    and Device parameters and internal variables on one hand, and the variables
    in the model, transistor, and 'geometry' data structures (see below) on the
    other.
    If data structures are already predefined for a particular implementation
    in a simulator, this table may be adapted to have the aliases point
    to the appropriate data locations to replace the Model->... etc. data
    structures.
*/


/**  EKV  intrinsic model parameters  **/

#define  MOS_TYPE Model->mostype
/*  Process related parameters  */
#define  XJ      Model->xj       /*  metallurgical junction depth            */
#define  DW      Model->dw       /*  channel narrowing                       */
#define  DL      Model->dl       /*  channel shortening                      */

/*  EKV Model basic parameters  */
#define  COX     Model->cox      /*  gate oxide capacitance                  */
#define  VTO     Model->vto      /*  nominal threshold voltage               */
#define  GAMMA   Model->gamma    /*  body effect parameter                   */
#define  PHI     Model->phi      /*  bulk Fermi potential                    */
#define  KP      Model->kp       /*  transconductance parameter              */
#define  E0      Model->e0       /*  vertical critical field  (ekv v > 2.5)         */
#define  EO      Model->e0       /*  !!  mb  98/08/08  */
#define  UCRIT   Model->ucrit    /*  longitudinal critical field             */

/*  !!  mb  98/08/08   Optional parameters  */
#define  TOX     Model->tox      /*  gate oxide thickness (optional)         */
#define  VFB     Model->vfb      /*  flat-band voltage (optional)            */
#define  NSUB    Model->nsub     /*  substrate doping (optional)             */
#define  UO      Model->uo       /*  low-field mobility (optional)           */
#define  VMAX    Model->vmax     /*  saturation velocity (optional)          */
#define  THETA   Model->theta    /*  mobility reduction coefficient (reinstated, optional)  */

/*  Fine tuning parameters  */
#define  LAMBDA  Model->lambda   /*  depletion length coefficient            */
#define  WETA    Model->weta     /*  narrow channel effect coefficient       */
#define  LETA    Model->leta     /*  short channel effect coefficient        */

/*  Substrate current parameters  */
#define  IBA     Model->iba      /*  !!  mb  95/12/19   */
#define  IBB     Model->ibb      /*  !!  mb  95/12/19   */
#define  IBN     Model->ibn      /*  !!  mb  95/12/19   */

/*  Reverse short channel effect parameters  */
#define  Q0      Model->q0       /*  !!  mb  97/04/14   */
#define  QO      Model->q0       /*  !!  mb  98/08/08   */
#define  LK      Model->lk       /*  !!  mb  97/04/14   */

/*  Temperature parameters  */
#define  TCV     Model->tcv      /*  threshold voltage temp. coefficient     */
#define  BEX     Model->bex      /*  mobility temperature exponent           */
#define  UCEX    Model->ucex     /*  long. crit. field temperature exponent  */
#define  IBBT    Model->ibbt     /*  !!  mb  95/12/19   */

/*  Short distance matching parameters (EKV v2.6)  !!  mb  97/07/18  */
#define  AVTO    Model->avto            /*     */
#define  AGAMMA  Model->agamma          /*     */
#define  AKP     Model->akp             /*     */

/*  Noise parameters  */
#define  KF      Model->kf       /*  flicker noise coefficient               */
#define  AF      Model->af       /*  flicker noise exponent                  */

/*  Setup parameters  */
#define  NQS     Model->nqs      /*  nonquasistatic operation coefficient    */
#define  SATLIM  Model->satlim   /*  if/ir ratio saturation limit            */
#define  EKVINT  Model->ekvint   /*  !!  mb  96/08/23  added interpolation function selector  */
#define  EKVDYN  Model->ekvdyn   /*  !!  mb  97/07/18  added choice for zero capacitances  */
#define  XQC     Model->xqc      /*  !!  mb  97/07/18  added choice for zero capacitances  */
#define  SCALEM  Model->scalem   /*  model scaling factor                    */
#define  UPDATE  Model->update   /*  !!  mb  97/06/02  model version selector  */


#define  CGSO    Model->cgso     /*  gate-source overlap capacitance         */
#define  CGDO    Model->cgdo     /*  gate-drain overlap capacitance          */
#define  CGBO    Model->cgbo     /*  gate-bulk overlap capacitance           */


/*  Temperature dependent model variables  */
#define  Temp    Model->temp     /*  model temperature                       */
#define  Vt      Model->ktonq    /*  thermal voltage                         */
#define  oldVt   Model->oldktonq
#define  Eg      Model->egfet    /*  silicon bandgap energy                  */
#define  oldEg   Model->oldegfet

/*  Geometry & voltage independent variables (internal to the EKV model)  */

#define  eps_COX         Model->eps_cox
#define  Lc              Model->lc
#define  Lc_LAMBDA       Model->lc_lambda

#define  Vt_2            Model->vt_2
#define  Vt_4            Model->vt_4             /*  !!  mb  95/12/19  */
#define  Vt_Vt           Model->vt_vt
#define  Vt_Vt_2         Model->vt_vt_2
#define  Vt_Vt_16        Model->vt_vt_16
#define  Vt_01           Model->vt_01            /*  !!  mb  96/10/28  */
#define  inv_Vt          Model->inv_vt           /*  !!  mb  97/04/21  */

#define  sqrt_PHI        Model->sqrt_phi
#define  deltaVGprime    Model->deltavgp         /*  !!  mb  09/09/99  */

#define  Lc_UCRIT        Model->lc_ucrit
#define  inv_UCRIT       Model->inv_ucrit

#define  IBN_2           Model->ibn_2            /*  !!  mb  96/07/15  */

#define  IBA_IBB         Model->iba_ibb          /*  !!  mb  95/12/19  */
#define  Lc_IBB          Model->lc_ibb           /*  !!  mb  95/12/19  */

#define  V0              Model->v0       /*  !!  mb  97/04/14   */

/* version 2.6 new parameters !!  mb  25/05/97  */
#define  T0              Model->t0          /*  !!  mb  97/06/02  ekv v2.6  */
#define  eta_qi          Model->e_qi        /*  !!  mb  97/06/02  ekv v2.6  */


/*  Geometry dependent & voltage independent variables  */

#define  Cgsov   Geo->cgsov
#define  Cgdov   Geo->cgdov
#define  Cgbov   Geo->cgbov

#define  Vc        Geo->gvc          /*  !!  mb  95/05/10  */
#define  WLCox     Geo->gwlcox
#define  log_Vc_Vt Geo->log_vc_vt    /*  !!  mb  95/05/10  */
#define  LETA_L    Geo->leta_l
#define  WETA_W    Geo->weta_w
#define  AWL       Geo->awl      /*  !!  mb  97/07/18  short-distance matching  */
#define  TAU       Geo->tau

/*  Transistor structure parameters  */

#define  L       Trans->l        /*  channel length                     */
#define  W       Trans->w        /*  channel width                      */
#define  Leff    Trans->leff     /*  effective channel length           */
#define  Weff    Trans->weff     /*  effective channel width            */

#define  AD      Trans->ad       /*  drain diffusion area               */
#define  AS      Trans->as       /*  source diffusion area              */
#define  PD      Trans->pd       /*  drain diffusion perimeter          */
#define  PS      Trans->ps       /*  source diffusion perimerter        */
#define  NRD     Trans->nrd      /*  relative drain resistivity         */
#define  NRS     Trans->nrs      /*  relative source resistivity        */
#define  RDC     Trans->rdc      /*  drain contact resistance           */
#define  RSC     Trans->rsc      /*  source contact resistance          */

#define  NP      Trans->np       /*  multiple parallel device number    */
#define  NS      Trans->ns       /*  multiple series device number      */

#define  M       Trans->np       /*  synonym for NP  !! mb 99/03/26   */
#define  N       Trans->ns       /*  synonym for NS  !! mb 99/03/26  */




/****************************************/
/*            DATA STRUCTURES           */
/****************************************/

/*------------------------------------------------------------------------------*/
/*  The Data structures below should help distinguish between variables to be
    computed depending on geometry and/or bias and/or temperature. Computation
    of these variables is grouped respectively in procedures, which have to be
    called from the simulator in an appropriate sequence according to its needs.
*/

/*------------------------------------------------------------------------------*/
/*  MOS Model structure which contains all the parameters which can be specified
    in the .MODEL card. Additionnally, working variables valid for all devices
    of that specific model are stored.
    The associated global variable is '*mosmodels'. In every transistor
    component structure, 'imodels' indicates the MOS model to be used in the
    mosmodels table, which is accessed with 'Model->...'.
    It is useful to keep in mind that future extensions of the model may need
    an adaptation of the data structures.
*/

typedef  struct  s_typemodel
{
/*  Intrinsic model parameters  */

    /*  EKV Model basic parameters  */
    double      cox, vto, gamma, phi, kp, e0, theta, ucrit;  /*  !!  mb  97/06/02   */
    double      nfs, xj;
    double      dl, dw;

    /*  Optional process parameters for calculation of electrical parameters  */
    /*  Note: units of TOX: m, NSUB: cm^^-3, VFB: V; UO: cm^^2/(Vs)           */
    double      tox, nsub, vfb, uo, vmax;             /*  !!  mb  98/08/08  */

    /*  Fine tuning & setup parameters  */
    double      lambda, weta, leta;
    double      nqs, satlim, ekvint, ekvdyn, xqc, update, scalem;  /*  !!  mb  97/06/02  added ekvint, update,... */

    /*  Impact ionization parameters  */
    double      iba, ibb, ibn;                        /*  !!  mb  99/05/10  removed ibc */

    /*  RSCE model parameters  (EKV v2.5)  */
    double      q0, lk;                               /*  !!  mb  97/04/14   */

    /*  Temperature parameters & variables  */
    double      tcv, bex, ucex, ibbt;                 /*  !!  mb  95/12/19   added ibbt */
    double      temp, oldtemp, tnom;

    /*  Short distance matching parameters  */
    double      avto, agamma, akp;                    /*  !!  mb  97/07/18  */

    /*  Noise model parameters  */
    double      kf, af;

/*  Extrinsic model parameters  */

    double      cgso, cgdo, cgbo;


/*  Other parameters & variables  */

    char        mostype;       /*  'N' or 'P'  */

    /*  'working variables'  */
    double      ktonq, oldktonq, egfet, oldegfet;
    double      eps_cox, lc, lc_lambda;

    /*  !!  mb  95/05/10, 95/12/19  added vt_4, 96/10/28  added vt_01, ddv, 97/04/21  added inv_vt
                97/09/03  removed cox_kp_vt  */
    double      vt_2, vt_4, vt_vt, vt_vt_2, vt_vt_16, vt_01, ddv, inv_vt;

    double      phi_half, five_phi_half, inv_phi_sqrt;
    double      sqrt_phi, lc_ucrit, inv_ucrit;  /*  !!  mb  96/10/28,  97/07/18  */
    double      ibn_2;                          /*  !!  mb  96/07/15  */

    double      iba_ibb, lc_ibb;                /*  !!  mb  95/12/19  */

    double      v0;                             /*  !!  mb  97/04/14  */

    double      t0, e_qi;                       /*  !!  mb  97/06/02  ekv v2.6  */
    double      deltavgp;                       /*  !!  mb  99/09/09  */
} typemodel;


/*------------------------------------------------------------------------------*/
/*  Transistor structure used for storage of the instance parameters:

      nametrans  n2  ng  n1  nb  imodel  w  l  [wd  ld  ad  as  pd  ps  nrd  nrs]

      nametrans           name of the transistor
      n2 ng n1 nb         topological nodes of the transistor
      imodel              index of the typemodel structure in the mosmodels table
                          to which belongs the component

    For each transistor, one typetrans structure is allocated. The associated
    global variable which contains the various transistor structures is
    '*mostrans'; the components' variables are accessed with 'Trans->...'.
    Note that for certain implementations, it is preferred to store the electrical
    channel lengths and widths (leff, leff) in the variables l, w to save place.
*/

typedef  struct  s_typetrans
{
  long    imodel;
    double      w, l, weff, leff;
    double      ad, as, pd, ps;
    double      nrd, nrs;
    double      np, ns;
    double      rdc, rsc;
    long        ng, n1, n2, nb;           /*  Transistor nodes  */
    char        * nametrans;
} typetrans;


/*------------------------------------------------------------------------------*/
/*  For each transistor, one typegeo structure is allocated. It contains
    variables which are used to store values which do not change with
    bias, but which depend on geometries of the component.
    The associated global variable which contains the various geo structures
    is '*mosgeo', the variables are accessed with 'Geo->...'.
    Note that most variables only depending on the geometry of the device are
    located in this structure; they are generally computed only once for any
    device.
*/


typedef  struct  s_typegeo
{
    double      c1b, c2b, cg1, cg2, cgb;
    double      cgsov, cgdov, cgbov;
    double      gn;          /*  !!  mb  98/07/09  for th. noise computation; rem. if_, ir  */
    double      gm;          /*  !!  mb  98/07/09  for fl. noise computation  */
    double      tau;         /*  !!  mb  97/07/18  for nqs computation  97/09/03  rem. tau_zero   */
    double      gwlcox, kp_weff, log_vc_vt, gvc;    /*  !!  mb  95/05/10  */
    double      leta_l, weta_w;
    double      awl;                  /*  !!  mb  97/07/18  for short-distance matching  */
} typegeo;


/*------------------------------------------------------------------------------*/
/*  Field for non-quasi-static model, for AC only
*/

typedef  struct  s_typenqsfield
{
	double      gdd, gdg, gds, gsd, gsg, gss, gbd, gbs;
} typenqsfield;


/************************************/
/*         GLOBAL  DATA             */
/************************************/

struct model_global_data {
  /*  Models and transistor data structures  */
  typemodel   *mosmodels;
  typetrans   *mostrans;
  typegeo     *mosgeo;
  double      eps;
  double      current;
  double      dids_dvg1, dids_dvb1, dids_dv21;
  double      ib1, ib2;
  double      dib1_dvgb, dib1_dv2b, dib1_dv1b;
  double      dib2_dvgb, dib2_dv2b, dib2_dv1b;
  double      gmb1, gmb2;
  double      cgb, cg1, cg2, c1b, c2b;
  double      *memccurrent, *ccurrent;
  double      ia, va1, va2, va3;
  double      ktonq, egfet, oldegfet;

  /*  Default values for MOS geometries specified in the .OPTION statement  */
  double      defw, defl, defpd, defps, defad, defas, defnrd, defnrs;

  /*  Execution control flags. These need to be set by the simulator into
      which the code is implemented according to the situation of analysis
  */
  char        access_res;             /*  presence of access resistors        */
  char        analytic_deriv;         /*  analytical derivatives computation  */
  char        newton_analysis;        /*  Newton analysis                     */
  char        transient_analysis;     /*  transient analysis                  */
  char        ac_analysis;            /*  ac analysis                         */

  char        operating_point;        /*  operating point                     */
  char        inv_polar;              /*  inverse polarisation                */
};

typedef struct model_global_data model_global_data_t;


/************************************/
/*         FUNCTION PROTOTYPE        */
/************************************/

static void F_v_n(double, double *, double *,double * );
static void F_v_simple (double, double *, double *, double *);

void ekv_initdefaultgeo(model_global_data_t * global, long);
void ekv_initgeo(model_global_data_t * global, long);

void ekv_initmodelparams(model_global_data_t * global, long);    // testowanie poprawnosci parametrow
void ekv_initmodel(model_global_data_t * global, long);

double ekv_moseval(model_global_data_t * global, unsigned int itrans, double vng, double vn1, double vn2,double  vnb);	// oblicznie pr�du IDS




/*************************************************************************/
/*          F(x) interpolation function and derivative:                  */
/*************************************************************************/

/*  !!  mb  95/08/28  EKV code version 2.2

    This new 'true' interpolation function replaces the simplified,
    approximate or tabulated versions used before.

    To speed up computation, the current, its square root, and the
    derivative are put together in one function.

    !!  mb  95/12/19  modified names for int. function constants

    !!  mb  97/02/10  ekv v2.4 interpolation function

*/



/*************************************************************************/
static void F_v_simple (double v, double *i_v, double *sqrt_i_v, double *d_i_v)
//double v;
//double *i_v, *sqrt_i_v, *d_i_v;

/*  !!  mb  96/08/23  introduced simple interpolation function as an option  */
{
     double  u = 0.5*v;

     if (u > 20.0)
     {
       *sqrt_i_v = u;
       *i_v = u*u;
       *d_i_v = u;
     }
     else if (u > -18.0)
     {
       double exp_u = exp(u);
       double s_i_v = log(1.0+exp_u);

       *sqrt_i_v = s_i_v;
       *i_v = s_i_v*s_i_v;
       *d_i_v = exp_u*s_i_v/(1.0+exp_u);
     }
     else
     {
       double exp_u = exp(u) + 1.0e-32;
       double exp_u2 = exp_u*exp_u;

       *sqrt_i_v = exp_u;
       *i_v = exp_u2;
       *d_i_v = exp_u2;
     }
}

/*************************************************************************/
static void F_v_n(double v, double *i_v, double *sqrt_i_v,double * d_i_v)
//double v;
//double *i_v, *sqrt_i_v, *d_i_v;

/*  !!  mb  97/02/10  ekv v2.4 interpolation function  */
{
     double  yk;

     if (v > -0.35)
     {
       double  z0 = 2.0/(1.3+v-log(v+1.6));
       double  zk = (2.0+z0)/(1.0+v+log(z0));
       yk = (1.0+v+log(zk))/(2.0+zk);
     }
     else if (v > -15.0)
     {
       double  z0 = 1.55+exp(-v);
       double  zk = (2.0+z0)/(1.0+v+log(z0));
       yk = (1.0+v+log(zk))/(2.0+zk);
     }
     else if (v > -23.0)
       yk = 1.0/(2.0+exp(-v));
     else
       yk = exp(v) + 1.0e-64;

     *i_v = yk*(1.0+yk);
     *sqrt_i_v = sqrt(*i_v);
     *d_i_v = yk;

}


/*************************************/
/*        EKV FUNCTIONS              */
/*************************************/

/*  NOTE:  The code is distributed into various functions which need to be called
           respectively when a bias, geometry or temperature update needs to be
           done; this way a minimum effort can be ensured for model computation.
           It should be relatively easy to find out the precedence of the various
           functions.
*/


/***********************************************************************/
/*
    Attribution of default values for model parameters. It is assumed
    that the model parameters have been read by some other procedure,
    and unassigned parameters given a value below the constant 'undefined'.
    Some parameter range tests are performed and default values are
    given for unassigned parameters or when they are out of (mathe-
    matical) range.

    Note: as PMOS are handled as pseudo-NMOS (see also in ekv_moseval),
    the sign of VTO and TCV is inverted for PMOS (thus VTO and TCV are
    supposed to be generally negative for PMOS).

    IMPORTANT:  The only place where scaling is needed for the EKV
    model is the expression 'epssil*SCALEM' (see below), if ALL model
    AND device parameters are specified with the same length units base
    (i.e. m or um). Thus default values for model parameters depending
    on the length units must be given the corresponding units.
*/

//void  ekv_initmodelparams(imodel)
//  long    imodel;

void  ekv_initmodelparams(model_global_data_t * global, long imodel)
{
  typemodel  *Model;

  double  tmp1, tmp2;                  /*  !!  mb  99/07/30  */

  Model = global->mosmodels + imodel;


/*  EKV Model setup parameters  */

  if (isnan(SCALEM) || SCALEM <= 0.0)  SCALEM = 1.0;
  UPDATE = 2.6;                        /*  !!  mb  98/08/08  UPDATE handling removed  */
  if (EKVINT != 1.0)  EKVINT = 0.0;    /*  interpolation function option, EKVINT = 1 chooses F(v) = ln^2(1+exp(v/2))  */
  if (EKVDYN != 0.0)  EKVDYN = 1.0;    /*  EKVDYN = 0 sets intr. capacitances to zero (no dynamic model)  */

  if (isnan(XQC) || XQC < 0.5)                  /*  !!  mb  99/05/10  (cd)  */
    XQC = 0.4;
  else
    XQC = 1.0;


/*  EKV Model basic parameters  */

  /*  Junction depth                        */
  if (isnan(XJ))             /*  !!  mb  98/08/26  (rg)  */
    XJ = 0.1e-6/SCALEM;
  else if (XJ < 1.0e-9)
    XJ = 1.0e-9/SCALEM;

  /*  Channel Length correction             */
  if (isnan(DL))             /*  !!  mb  00/10/31  */
    DL = 0.0;

  /*  Channel Width correction              */
  if (isnan(DW))             /*  !!  mb  00/10/31  */
    DW = 0.0;


  /*  !!  mb  98/08/08  introduced optional parameters             */
  /*  Note: units of TOX: m, NSUB: cm^^-3,  VFB: V,  UO: cm^^2/(Vs),  VMAX: m/s  */

  /*  Gate oxide capacitance                */
  if (isnan(COX))
  {
    if (!isnan(TOX) && TOX > 0.0)           /*  !!  mb  98/08/08  */
    {
      /*  Cox = Eox / Tox  */
      COX = epsox*SCALEM / TOX;
    }
    else
	  COX = 0.7e-3*SCALEM*SCALEM;               /*  !!  mb  99/05/10  (cd)  */

  }
  else if (COX <= 0.0)
    COX = 0.7e-3*SCALEM*SCALEM;


  /*  Body effect parameter                 */
  if (isnan(GAMMA))
  {
    if (NSUB >= 0.0)           /*  !!  mb  98/08/26  */
    {
      /*  Gamma = sqrt(2qEsi*Nsub)/Cox  */
      GAMMA = sqrt(twoq_esi*(NSUB*1.0e6)) / COX;   /*  !!  mb  98/08/28  corrected  */
    }
    else
      GAMMA = 1.0;
  }
  else if (GAMMA < 0.0)
    GAMMA = 0.0;


  /*  Bulk Fermi potential                  */
  /*  bandgap  */
  Eg = 1.16 - (7.02e-4*Temp*Temp) / (Temp + 1108.0);
  /*  Vt = k*T/q;  */
  Vt = konq*Temp;
  if (isnan(PHI))
  {
    if (NSUB > 0.0)           /*  !!  mb  98/08/08  */
    {
      /*  ni = 1.45e10*(temp/reftmp)*exp(Egreftmp-0.5*Eg/Vt);  (in cm-3) */
      /*  Phi = 2.0*Vt*ln(Nsub/ni);                            */
      /*  !!  mb  98/08/28  corrected PHI(Tnom) code (jb)  */
      PHI = 2.0*Vt*log(NSUB*reftmp/(1.45e10*Temp)) - Egreftmp*(Temp/reftmp) + Eg;

    }
  }
  else if (PHI <= 0.0)
    PHI = 0.7;

  /*  if (PHI < 0.1)  PHI = 0.1;  */

  /*  !!  mb  99/07/30  prevents PHI from becoming smaller than 0.2  */
  tmp1 = 0.2;
  tmp2 = PHI - tmp1;
  PHI = 0.5*(tmp2 + sqrt(tmp2*tmp2 + Vt*Vt)) + tmp1;


  /*  Nominal threshold voltage             */
  /*  PMOS is handled as pseudo NMOS with inverted voltage signs (for VTO, VFB)  */
  if (isnan(VTO))
  {
    if (!isnan(VFB))           /*  !!  mb  98/08/08  */
    {
      if (Model->mostype == 'P')
        VFB = -VFB;
      VTO = VFB + PHI + GAMMA*sqrt(PHI);
    }
    else
      VTO = 0.5;
  }
  else
    if (Model->mostype == 'P')
      VTO = -VTO;


  /*  Transconductance parameter            */
  if (isnan(KP))
  {
    if (UO > 0.0)           /*  !!  mb  98/08/08  */
    {
      /*  Kp = Uo * Cox  */
      KP = (UO*1.0e-4) * COX;     /*  !!  mb  98/08/28  corrected (jb) */
    }
  }
  else if (KP <= 0.0)
    KP = 5.0e-5;


  /*  Longitudinal critical field           */
  if (isnan(UCRIT))
  {
    if (VMAX > 0)           /*  !!  mb  98/08/08  */
    {
      /*  Ucrit = Vmax/Uo  */
      UCRIT = VMAX*COX/KP;           /*  !!  mb  98/08/26  corrected (jb) */
    }
    else
      UCRIT = 2.0e6*SCALEM;
  }
  else if (UCRIT <= 0.0)
    UCRIT = 100.0e6*SCALEM;       /*  disable UCRIT  */


  /*  Mobility reduction                    */
  if (isnan(E0))         /*  !!  mb  98/08/08  */
  {
    if (!isnan(THETA))
      E0 = 0.0;                   /*  simple mobility model is used  */
    else
      E0 = 1.0e12*SCALEM;         /*  disable E0    */
  }
  else if (E0 <= 0.0)
    E0 = 1.0e12*SCALEM;           /*  disable E0    */

  if (isnan(THETA) || THETA < 0.0)  THETA = 0.0;  /*  ensure that THETA >= 0 in all cases  */


  /*  Fine tuning coefficients  */
  if (isnan(LAMBDA))  LAMBDA = 0.5;                 /*  depletion length coefficient          */
  if (isnan(WETA) || WETA < 0.0)  WETA = 0.25;                          /*  narrow channel effect coefficient     */
  if (isnan(WETA) || LETA < 0.0)  LETA = 0.1;                           /*  short channel effect coefficient      */

  /*  Impact ionization parameters  */                   /*  !!  mb  95/12/19  */
  if (isnan(WETA) || IBA < 0.0)  IBA = 0.0;                             /*  1st impact ionization parameter       */
  if (isnan(WETA) || IBB < 0.0)  IBB = 3.0e8*SCALEM;                    /*  2nd impact ionization parameter       */
  if (isnan(IBN) || IBN < 0.0)  IBN = 1.0;                             /*  3rd impact ionization parameter       */

  /*  RSCE parameters  !!  mb  97/04/14 */
  if (isnan(IBN) || Q0 < 0.0)      Q0 = 0.0;            /*  RSCE excess charge  (usual value 3.116e-4 [As])     */
  if (isnan(LK) || LK <= 0.0)     LK = 2.9e-7/SCALEM;  /*  RSCE characteristic length (default is usual [m])   */

  /*  Short distance matching parameter  !!  mb  25/05/97  */
  if (isnan(AVTO))   AVTO = 0.0;     /*  short-distance matching for VTO    */
  if (isnan(AGAMMA)) AGAMMA = 0.0;   /*  short-distance matching for GAMMA  */
  if (isnan(AKP))    AKP = 0.0;      /*  short-distance matching for KP     */

  /*  Temperature parameters (intrinsic)  */
  if (isnan(TCV))  TCV = 1.0e-3;     /*  threshold voltage temp coefficient                                   */
  else if (Model->mostype == 'P')         /*  PMOS is handled as pseudo NMOS with inverted voltage signs, see VTO  */
    TCV *= -1.0;                          /*  !!  mb  96/08/23  TCV is supposed to be negative for PMOS            */

  if (isnan(BEX))  BEX = -1.5;       /*  mobility temperature exponent                                        */
  if (isnan(UCEX))  UCEX = 0.8;      /*  UCRIT temperature exponent  !!  mb  95/12/19  changed default        */
  if (isnan(IBBT))  IBBT = 9.0e-4;

  /*  Noise parameters  */
  if (isnan(KF))  KF = 0.0;                         /*  flicker noise coefficient             */
  if (isnan(AF))  AF = 1.0;                         /*  flicker noise exponent                */

  /*  Setup coefficients  */
  if (NQS != 1.0)  NQS = 0.0;                            /*  nonquasistatic operation switch       */
  if (isnan(SATLIM) || SATLIM < 0.0)  SATLIM = 54.59815;                  /*  saturation limit if/ir, def. e(4)     */
                                                         /*  only needed for .OP info purposes     */

  /*  Overlap capacitances parameters  */
  if (isnan(CGSO) || CGSO < 0.0)  CGSO = 0.0;                           /*  gate-source overlap capacitance       */
  if (isnan(CGDO) || CGDO < 0.0)  CGDO = 0.0;                           /*  gate-drain overlap capacitance        */
  if (isnan(CGBO) || CGBO < 0.0)  CGBO = 0.0;                           /*  gate-bulk overlap capacitance         */


/*  Geometry, voltage and temperature independent model variables  */

  eps_COX = epssil*SCALEM/COX;                           /*  !!  mb  95/12/19 (cd)                 */
  Lc = sqrt(eps_COX*XJ);                                 /*  #define  Lc         Model->lc         */
  Lc_LAMBDA = Lc*LAMBDA;                                 /*  #define  Lc_LAMBDA  Model->lc_lambda  */

  IBN_2 = 2.0*IBN;                                       /*  !!  mb  96/07/15  #define  IBN_2      */

  V0 = (Q0+Q0) / COX;                                    /*  !!   mb  97/04/14  */

  /*  !!  mb  98/08/08  corrected so as to account for epssil scale  */
  /*  T0 = COX / (epssil*E0);  */
  if (E0 > 0.0)
    T0 = 1.0 / (eps_COX*E0);

  if (Model->mostype == 'N')
    eta_qi = 0.5;
  else
    eta_qi = one3rd;

}



/***********************************************************************/
/*  Temperature update of model parameters.
    Extrinsic elements temperature update is done as in conventional
    models, except for series resistances (quadratic model).

    Note: here, temperature is handled as uniform for all transistors of
    a given model. Certain implementations may allow to choose temperature
    independently for each transistor, which is not described here.
*/

/*  newtemp: current temperature, oldtemp: previous temperature  */

void  ekv_tempparams(model_global_data_t * global, long imodel, double oldtemp, double newtemp, double reftemp)
{
  double  refVt, refEg;
  double  ratio, logratio;
  double  refratio, logrefratio;
  double  pbratio;
  double  tmp1, tmp2;                  /*  !!  mb  99/07/30  */

  {
    register struct s_typemodel  *Model = &(global->mosmodels[imodel]);

    ratio = newtemp / oldtemp;
    logratio = log(ratio);
    refratio = reftemp / oldtemp;
    logrefratio = log(refratio);

    /*  Thermal voltage update  */
    oldVt = Vt;
    Vt = konq * newtemp;
    refVt = konq * reftemp;

    /*  Energy gap temperature update  */
    oldEg = Eg;
    Eg = 1.16 - (7.02e-4*newtemp*newtemp) / (newtemp + 1108.0);
    refEg = 1.16 - (7.02e-4*reftemp*reftemp) / (reftemp + 1108.0);


/*  Note that all of the following parameters are updated 'iteratively'
    i.e. based on the value for the preceding temperature
*/

/*  EKV intrinsic model parameters update  */

    /*  update VTO=VTO-TCV*(TEMP-TNOM)  */
    VTO = VTO - TCV * (newtemp - oldtemp);
/*  !!  mb  The above solution is applied for both N- and PMOS (pseudo NMOS),
            since at this point VTO > 0 and TCV > 0 (see parameter initialisation;
            the temperature law similar for the normally positive value of VTO
			for NMOS and PMOS: |VTO| decreases for higher temperature)
*/

    /*  update KP=KP*(T/Tref)^BEX  */
    KP *= exp(BEX * logratio);

    /*  update THETA=THETA*sqr(Tref/T)  */
/*  !!   mb  THETA update in temperature does not seem to be correct;
             no temperature dependency is found in the literature
    THETA *= exp(-2.0 * logratio);
*/

    /*  update UCRIT=UCRIT*(T/Tref)^UCEX  */
    UCRIT *= exp(UCEX * logratio);

    /*  update PHI(T)=PHI*T/Tref - 3*Vt*ln(T/Tref) - Eg(Tref)*T/Tref + Eg(T)  */
    PHI = (PHI - oldEg)*ratio - 3.0*Vt*logratio + Eg;

    /*  !!  mb  99/07/30  prevents PHI from becoming smaller than 0.2  */
    tmp1 = 0.2;
    tmp2 = PHI - tmp1;
    PHI = 0.5*(tmp2 + sqrt(tmp2*tmp2 + Vt*Vt)) + tmp1;


    /*  !!  mb  95/12/19  update IBB(T) = IBB(Tref)*(1+IBBT*(T-Tref)  */
    IBB *= (1.0 + IBBT*(newtemp-reftemp))/(1.0 + IBBT*(oldtemp-reftemp));


/*  Flicker noise coefficient  */
/*  !!  mb  96/06/19   excluded from code
    AF = AF * pbratio;
*/
 }
}



/*******************************************************************/
/*  Initialization of model working variables depending on temperature
    updated model parameters
*/
void  ekv_initmodel(model_global_data_t * global, long imodel)
{
  typemodel  *Model;

  double     tmp1;

  Model = global->mosmodels + imodel;

/*  Model working variables, geometry and voltage independent, which need to be
    initialized at start, and updated after temperature change
*/

  /*  EKV model internal variables depending on temperature  */
  Vt_2 = 2.0*Vt;                        /*  Model->vt_2                  */
  Vt_4 = 4.0*Vt;                        /*  Model->vt_4   !!  mb  96/04/12  (gg)  */
  Vt_Vt = Vt*Vt;                        /*  Model->vt_vt                 */
  Vt_Vt_2 = 2.0*Vt_Vt;                  /*  Model->vt_vt_2               */
  Vt_Vt_16 = 16.0*Vt_Vt;                /*  Model->vt_vt_16              */
  inv_Vt = 1.0/Vt;                      /*  Model->inv_vt  !!  mb  97/04/21  */

  Vt_01 = 0.1*Vt;                         /*  !!  mb  96/10/28  Model->vt_01       */

  /*  !!  mb 95/05/10 replaced inv... by Lc_UCRIT  */
  Lc_UCRIT = Lc*UCRIT;                  /*  Model->lc_ucrit              */

  /*  !!  mb 97/07/18  introduced  inv_UCRIT, sqrt_PHI  */
  inv_UCRIT = 1.0/UCRIT;                /*  Model->inv_ucrit             */
  sqrt_PHI = sqrt(PHI);                 /*  Model->sqrt_phi              */

  /*  !!  mb  17/08/99  introduced deltaVGprime  */
  tmp1 = PHI+GAMMA*sqrt_PHI;
  deltaVGprime = 0.5*(-tmp1+sqrt(tmp1*tmp1+2.0*Vt_Vt_16));    /*  !!  mb  99/09/09  */

  Lc_IBB = Lc*IBB;                      /*  !!  mb  96/04/12  moved to ekv_initmodel (pm)  */
  IBA_IBB = IBA/IBB;                    /*  !!  mb  96/04/12  moved to ekv_initmodel (pm)  */


}



/*******************************************************************/
/*  Initialization of default geometries for each transistor.
*/
void  ekv_initdefaultgeo(model_global_data_t * global, long itrans)
{
  typetrans  *Trans;
  typemodel  *Model;

  Trans = global->mostrans + itrans;
  Model = global->mosmodels + Trans->imodel;

  if (W < 0.0) {
    if (global->defw < 0.0) W = 0.0;
    else W = global->defw;
  }

  if (L < 0.0) {
    if (global->defl < 0.0) L = 0.0;
    else L = global->defl;
  }

  /*  effective channel length and width  */
  Weff = NP * (W + DW);
  Leff = NS * (L + DL);

  if ((Weff <= 0.0) || (Leff <= 0.0))
    printf("Error: dimensions of transistor: %s is less than zero\n", Trans->nametrans);

}



/*****************************************************************/
/*  Computation of geometry dependent device variables.
    Temperature update is supposed to be done prior to this
    calculation.

*/

void  ekv_initgeo(model_global_data_t * global, long itrans)
{
  typetrans       *Trans;
  typegeo         *Geo;
  typemodel       *Model;

  Trans = global->mostrans + itrans;
  Geo = global->mosgeo + itrans;                    /*  !!  mb  95/05/14  corrected (sa)  */
  Model = global->mosmodels + Trans->imodel;

/*  Geometry dependent & voltage independent variables ('int. EKV')  */

  Vc = UCRIT*Leff;                    /*  NOTE: use L if necessary  */
  log_Vc_Vt = Vt*(log(0.5*Vc*inv_Vt)-0.6);   /*  !!  mb  98/02/05  (r1)  */
  WLCox = Weff * Leff * COX;

  AWL = 1.0/sqrt(Weff*Leff);       /*  !!  mb  97/07/18  short-distance matching  */

  /*  !!  mb  99/03/26  corrected for multiple device factors  */
  WETA_W = eps_COX*3.0*WETA*NP/Weff;
  LETA_L = eps_COX*LETA*NS/Leff;


  /*  Gate overlap capacitances  */

  Cgsov = Weff * CGSO;               /*  NOTE: use W, L if necessary  */
  Cgdov = Weff * CGDO;
  Cgbov = Leff * CGBO;

}


/***********************************************************************/
/*  !!  mb  98/07/09  changed function call  */

double  ekv_flickernoise(model_global_data_t * global, long itrans, double freq)
{
  typetrans      *Trans;
  typegeo        *Geo;
  typemodel      *Model;

  double  GM;

  Trans = global->mostrans + itrans;
  Geo = global->mosgeo + itrans;
  Model = global->mosmodels + Trans->imodel;

  GM = Geo->gm;

  return (KF * GM * GM / (COX * Leff * Weff * exp(AF * log(freq))));
}


/***********************************************************************/
/*  !!  mb  98/07/09  changed function call, updated  */

double  ekv_thermalnoise(model_global_data_t * global, long itrans, double fourkt)
{
  typetrans      *Trans;
  typegeo        *Geo;

  double  GN;

  Trans = global->mostrans + itrans;
  Geo = global->mosgeo + itrans;

  GN = Geo->gn;           /*  noise conductance  */

  /*  thnoise = 4kT*(mu/(Leff^^2))*|QI| = 4kT*beta*|qi|  */

  return (fourkt * GN);
}



/***********************************************************************/
/*                      EKV Model evaluation function:                 */
/***********************************************************************/

/*  This function computes the following:
      - drain-source current IDS from pins n2 to n1, using the
        respective voltages (current = f(vng, vn1, vn2, vnb)), and
        derivatives of the current with respect to the voltages
      - impact ionization current and associated conductances
      - intrinsic charges resp. capacitances
*/


/**
 * Structure and model parameters initialization.
 *
 * The first argument of this function is an array of parameters
 * values which are to be set in the model. The parameters are
 * given in a predefined order: cox, vto, gamma, phi, kp, e0, ucrit.
 *
 * @see setDefaultEKVParametersValues
 *
 * @param param_values Parameters' values to be set.
 * @param imodel
 * @param itrans
 *
 * @author S. Szostek
 */
void init_param(model_global_data_t * global, const double * param_values, long imodel, long itrans)
{
	// translate nan into par_not_def
	#define SET(param,value) if (isnan(value)) param = NAN; else param = value;

	typemodel  *Model = global->mosmodels + imodel;
	typetrans  *Trans = global->mostrans + itrans;
	typegeo    *Geo = global->mosgeo + itrans;

	Trans->imodel = imodel;
	// inicjowanie parametrow modelu
	MOS_TYPE = 'N';   //Model->mostype
/**  EKV  intrinsic model parameters  **/

/*  Process related parameters  */
	XJ  = 1.50E-07;    //Model->xj       /*  metallurgical junction depth            */
	DW  = -2.00E-08;    //Model->dw       /*  channel narrowing                       */
	DL  = -5.00E-08;    //Model->dl       /*  channel shortening                      */

/*  EKV Model basic parameters  */
	SET(COX,	param_values[6]);	//Model->cox      /*  gate oxide capacitance                  */
	SET(VTO,	param_values[0]);	//Model->vto      /*  nominal threshold voltage               */
	SET(GAMMA,	param_values[1]);	//Model->gamma    /*  body effect parameter                   */
	SET(PHI,	param_values[2]);	//Model->phi      /*  bulk Fermi potential                    */
	SET(KP,		param_values[3]);	//Model->kp       /*  transconductance parameter              */
	SET(UCRIT,	param_values[5]);	//Model->ucrit    /*  longitudinal critical field             */

	E0     = NAN;	//8.80E+07;				//Model->e0       /*  vertical critical field  (ekv v > 2.5)         */
	//	EO     = par_not_def;		//Model->e0       /*  !!  mb  98/08/08  */

/*  !!  mb  98/08/08   Optional parameters  */
	SET(THETA,	param_values[4]);	//Model->theta    /*  mobility reduction coefficient (reinstated, optional)  */
	SET(TOX,	param_values[7]); 	//25.0E-08;	Model->tox      /*  gate oxide thickness (optional)         */
	VFB    = NAN;		//-0.2; 	Model->vfb      /*  flat-band voltage (optional)            */
	NSUB   = NAN;		//2.0E+17;	Model->nsub     /*  substrate doping (optional)             */
	UO     = NAN;		//433.0; 	Model->uo       /*  low-field mobility (optional)           */
	VMAX   = NAN;		//1.8E+5;	Model->vmax     /*  saturation velocity (optional)          */

/*  Fine tuning parameters  */
    LAMBDA = 0.23;				//Model->lambda   /*  depletion length coefficient            */
    WETA   = 5.00E-02;			//Model->weta     /*  narrow channel effect coefficient       */
	LETA   = 2.80E-01;			//Model->leta     /*  short channel effect coefficient        */

/*  Substrate current parameters  */
	IBA    = 2.00E+08; //Model->iba      /*  !!  mb  95/12/19   */
	IBB    = 3.50E+08; //Model->ibb      /*  !!  mb  95/12/19   */
    IBN    = 1.0; //Model->ibn      /*  !!  mb  95/12/19   */

/*  Reverse short channel effect parameters  */
    Q0     = 2.80E-04; //Model->q0       /*  !!  mb  97/04/14   */
   // QO     = NAN; //Model->q0       /*  !!  mb  98/08/08   */
    LK     = 5.00E-07; //Model->lk       /*  !!  mb  97/04/14   */

/*  Temperat/ure parameters  */
	TCV    = 1.50E-03; //Model->tcv      /*  threshold voltage temp. coefficient     */
	BEX    = -1.5; //Model->bex      /*  mobility temperature exponent           */
	UCEX   = 1.7; //Model->ucex     /*  long. crit. field temperature exponent  */
	IBBT   = 0.00E+00; //Model->ibbt     /*  !!  mb  95/12/19   */

/*  Short distance matching parameters (EKV v2.6)  !!  mb  97/07/18  */
    AVTO    = 1.00E-08; //Model->avto            /*     */
	AGAMMA  = 1.00E-08;//Model->agamma          /*     */
    AKP     = 2.50E-08;//Model->akp             /*     */

/*  Noise parameters  */
	KF      = 1.00E-27;//Model->kf       /*  flicker noise coefficient               */
	AF      = 1.0;//Model->af       /*  flicker noise exponent                  */

/*  Setup parameters  */
	NQS     = NAN;//Model->nqs      /*  nonquasistatic operation coefficient    */
	SATLIM  = NAN;//Model->satlim   /*  if/ir ratio saturation limit            */
	EKVINT  = 0.0;//Model->ekvint   /*  !!  mb  96/08/23  added interpolation function selector  */
	EKVDYN  = NAN;//Model->ekvdyn   /*  !!  mb  97/07/18  added choice for zero capacitances  */
	XQC     = 0.0;//Model->xqc      /*  !!  mb  97/07/18  added choice for zero capacitances  */
	SCALEM  = NAN;//Model->scalem   /*  model scaling factor                    */
    UPDATE  = 2.6;//Model->update   /*  !!  mb  97/06/02  model version selector  */


	CGSO    = NAN;//Model->cgso     /*  gate-source overlap capacitance         */
	CGDO    = NAN;//Model->cgdo     /*  gate-drain overlap capacitance          */
	CGBO    = NAN;//Model->cgbo     /*  gate-bulk overlap capacitance           */


/*  Temperature dependent model variables  */
	Temp    = 298.0;//Model->temp     /*  model temperature                       */
	Vt      = NAN;//Model->ktonq    /*  thermal voltage                         */
    oldVt   = NAN;//Model->oldktonq
	Eg      = NAN;//Model->egfet    /*  silicon bandgap energy                  */
    oldEg   = NAN;//Model->oldegfet

/*  Geometry & voltage independent variables (internal to the EKV model)  */

    eps_COX     = NAN;    //Model->eps_cox
	Lc          = NAN;    //Model->lc
    Lc_LAMBDA   = NAN;    //Model->lc_lambda

    Vt_2        = NAN;    //Model->vt_2
	Vt_4        = NAN;    //Model->vt_4             /*  !!  mb  95/12/19  */
	Vt_Vt       = NAN;    //Model->vt_vt
	Vt_Vt_2     = NAN;    //Model->vt_vt_2
	Vt_Vt_16    = NAN;    //Model->vt_vt_16
	Vt_01       = NAN;    //Model->vt_01            /*  !!  mb  96/10/28  */
    inv_Vt      = NAN;    //Model->inv_vt           /*  !!  mb  97/04/21  */

    sqrt_PHI      = NAN;  //Model->sqrt_phi
	deltaVGprime  = NAN;  //Model->deltavgp         /*  !!  mb  09/09/99  */

	Lc_UCRIT      = NAN;  //Model->lc_ucrit
    inv_UCRIT     = NAN;  //Model->inv_ucrit

    IBN_2         = NAN;  //Model->ibn_2            /*  !!  mb  96/07/15  */

    IBA_IBB       = NAN;  //Model->iba_ibb          /*  !!  mb  95/12/19  */
	Lc_IBB        = NAN;  //Model->lc_ibb           /*  !!  mb  95/12/19  */

	V0            = NAN;  //Model->v0       /*  !!  mb  97/04/14   */

/* version 2.6 new parameters !!  mb  25/05/97  */
	T0            = NAN;  //Model->t0          /*  !!  mb  97/06/02  ekv v2.6  */
	eta_qi        = NAN;  //Model->e_qi        /*  !!  mb  97/06/02  ekv v2.6  */


/*  Geometry dependent & voltage independent variables  */

	Cgsov   = NAN;//Geo->cgsov
	Cgdov   = NAN;//Geo->cgdov
	Cgbov   = NAN;//Geo->cgbov

	Vc        = NAN;//Geo->gvc          /*  !!  mb  95/05/10  */
	WLCox     = NAN;//Geo->gwlcox
	log_Vc_Vt = NAN;//Geo->log_vc_vt    /*  !!  mb  95/05/10  */
	LETA_L    = NAN;//Geo->leta_l
	WETA_W    = NAN;//Geo->weta_w
	AWL       = NAN;//Geo->awl      /*  !!  mb  97/07/18  short-distance matching  */
	TAU       = NAN;//Geo->tau

/*  Transistor structure parameters  */

	SET(L,	param_values[8]);//Trans->l        /*  channel length                     */
	SET(W,	param_values[9]);//Trans->w        /*  channel width                      */
	Leff    = NAN;//Trans->leff     /*  effective channel length           */
	Weff    = NAN;//Trans->weff     /*  effective channel width            */

	AD      = NAN;//Trans->ad       /*  drain diffusion area               */
	AS      = NAN;//Trans->as       /*  source diffusion area              */
	PD      = NAN;//Trans->pd       /*  drain diffusion perimeter          */
	PS      = NAN;//Trans->ps       /*  source diffusion perimerter        */
	NRD     = NAN;//Trans->nrd      /*  relative drain resistivity         */
	NRS     = NAN;//Trans->nrs      /*  relative source resistivity        */
	RDC     = 0.1;//Trans->rdc      /*  drain contact resistance           */
	RSC     = 0.1;//Trans->rsc      /*  source contact resistance          */

	NP      = 1.0;//Trans->np       /*  multiple parallel device number    */
	NS      = 1.0;//Trans->ns       /*  multiple series device number      */

	M       = 1.0;//Trans->np       /*  synonym for NP  !! mb 99/03/26   */
	N       = 1.0;//Trans->ns       /*  synonym for NS  !! mb 99/03/26  */

  // inicjowanie parametr�w struktury


}

//-----------------------------
// Funkcja obliczaj�ca pr�d - modyfikacja funkcji oryginalnej tylko w zakresie pobierania
// parametr�w modelu i zwracania wynik�w oblicze�- nie ma zmian w samym sposobie obliczania pradu

double ekv_moseval(model_global_data_t * global, unsigned int itrans, double vng, double vn1, double vn2,double  vnb)
//double  vng, vn1, vn2, vnb;
{
  double  v1, v2;

  double  VD, VS, VG;             /*  intrinsic voltages ref. to bulk  */

/*  !!  mb  95/12/19  introduced variables for impact ionization, and
					  v2.3 related modifications
	!!  mb  95/08/29  introduced call to new interpol. fct., associated vars.
	!!  mb  95/05/11  rearranged variables precedence,
					  introduced new variables, suppressed old ones
*/

  double  VGprime, GAMMAprime;    /*  short and narrow channel effect  */
  double  VP, VPprime;            /*  pinch-off voltage                */
  double  if_;                    /*  forward normalized current       */
  double  VDSS, VDSSprime;        /*  saturation voltage  !! 95/05/10  */
  double  irprime;                /*  reverse normalized current !! 95/05/10  */
  double  deltaL, Leq;            /*  channel length reduction   !! 95/05/10  */
  double  beta;                   /*  transconductance factor          */
  double  n;                      /*  slope factor                     */
  double  Ispec, IDS;             /*  specific current, drain current  */

  double  sqrt_PHI_VD, sqrt_PHI_VS;
  double  big_sqrt_VP, sqrt_VP_Vt;                 /*  !!  mb  95/05/10  */
  double  sqrt_PHI_VP0, big_sqrt_VP0, VP0;         /*  !!  mb  97/09/14  */
  double  Vt_Vc;                                   /*  !!  mb  97/09/14  */
  double  ratioV_if, sqrt_if, dif_dv;              /*  !!  mb  95/08/29  */
  double  Vds, deltaV_2, Vip;                      /*  !!  mb  95/05/10  */
  double  Vdsprime;                                /*  !!  mb  97/07/18  */
  double  VDSS_sqrt, sqrt_VDSS_deltaV, sqrt_Vds_VDSS_deltaV;                 /*  !!  mb  95/05/10  */
  double  VDSSprime_sqrt, sqrt_VDSSprime_deltaV, sqrt_Vds_VDSSprime_deltaV;  /*  !!  mb  95/05/10  */
  double  ratioV_irprime, sqrt_irprime, dirprime_dv;                         /*  !!  mb  95/08/29  */
  double  Lprime, Lmin, sqrt_Lprime_Lmin;          /*  !!  mb  95/05/10  */
  double  THETA_VP_1, sqrt_PHI_VP;
  double  if_ir;

  double  GAMMAstar, sqrt_GAMMAstar;                       /*  !!  mb  95/12/19  */
  double  GAMMA_sqrt_PHI;                                  /*  !!  mb  97/07/18  */
  double  PHI_VD, PHI_VS, sqrt_PHI_VD_Vt, sqrt_PHI_VS_Vt;  /*  !!  mb  95/01/19  */

  double  Vib, inv_Vib, exp_ib, isub;                      /*  !!  mb  95/12/19  */
  double  Lc_IBB_Vib;                                      /*  !!  mb  96/06/20  */
  double  Isub, dIsub_factor;                              /*  !!  mb  96/09/03  */

  double  sqv, vL, deltaVFB;                               /*  !!  mb  97/04/14  */

  double  dQI_dVD, dQI_dVS, dQI_dVG;                       /*  !!  mb  97/06/02  */
  double  E0_Q_1, T0_GAMMA_1;

  double  VTOa, GAMMAa, KPa;                               /*  !!  mb  97/07/18  */


  /*  Variables used for derivatives computation  */
  double  dGAMMAprime_factor;                                  /*  !!  mb  95/12/19  */
  double  dGAMMAprime_dVD, dGAMMAprime_dVS;
  double  dGAMMAprime_dVG;                                     /*  !!  mb  97/09/14  */
  double  dVP_dVD, dVP_dVG, dVP_dVS;
  double  dVPprime_factor;
  double  dVPprime_dVD, dVPprime_dVG, dVPprime_dVS;
  double  dif_dVD, dif_dVG, dif_dVS;
  double  dVDSS_dVD, dVDSS_dVG, dVDSS_dVS;
  double  ddeltaV_dVD, ddeltaV_dVG, ddeltaV_dVS;
  double  inv_sqrt_VDSS_deltaV, inv_sqrt_Vds_VDSS_deltaV;      /*  !!  mb  97/04/21  */
  double  dVip_dVD, dVip_dVG, dVip_dVS;                        /*  !!  mb  95/05/10  */
  double  dVDSSprime_dVD, dVDSSprime_dVG, dVDSSprime_dVS;      /*  !!  mb  95/05/10  */
  double  dirprime_dVD, dirprime_dVG, dirprime_dVS;
  double  ddeltaL_dVD, ddeltaL_dVG, ddeltaL_dVS;
  double  dLeq_dVD, dLeq_dVG, dLeq_dVS;                        /*  !!  mb  95/05/10  */
  double  dbeta_dVD, dbeta_dVG, dbeta_dVS;
  double  dn_dVD, dn_dVG, dn_dVS;
  double  gds, gms, gmbs, gm;     /*  conductances                     */

  /*  Capacitances related variables  */
  double  ir, ratioV_ir, dir_dv;                           /*  !!  mb  95/05/10, 95/08/29  */
  double  sqrt_ir, sqrt_if_ir, quot_if_ir;
  double  cgss_if_ir, cgss_ir_if, cgbs;
  double  cgsw_if, cgsw_ir, cgbw;
  double  cgsi, cgdi, cgbi;
  double  Cgsi, Cgdi, Cgbi, Csbi, Cdbi; /*  intrinsic capacitances     */

  /*  Charges related variables  !!  mb  96/08/23  */
  double  sif, sir;
  double  sif2, sir2;
  double  sif3, sir3;
  double  sif_sir_2;
  double  QI, QD, QS, QB, QG;
  double  qi, qd, qs, qb, qg;                    /*  !!  mb  97/09/03  intrinsic charges  */

  double  n_Vt_COX, n_1, n_1_n, VP_PHI_eps, sqrt_PHI_VP_2;   /*  !!  mb  97/02/10 */

  double  sif_sir_3, dQ_i_factor;                            /*  !!  mb  96/11/28 */

  double  VGstar, sqrt_VGstar;                               /*  !!  mb  99/05/10 */

  double  sqrt_PHI_VP2_2;                                    /*  !!  mb  99/09/09 */

  double  dir_dVD, dir_dVS, dir_dVG;                         /*  !!  mb  97/02/10 */

  double  dQD_n_factor, d_QD_if, d_QD_ir;
  double  dQD_dVD, dQD_dVS, dQD_dVG;

  double  dQS_n_factor, d_QS_if, d_QS_ir;
  double  dQS_dVD, dQS_dVS, dQS_dVG;

  double  dQB_VP_factor;
  double  dQB_dVD, dQB_dVS, dQB_dVG;

  double  dQG_dVD, dQG_dVS, dQG_dVG;

  double  tau_zero;                                          /*  !!  mb  97/09/03 */

  double  VTH, VDSAT;                                        /*  !!  mb  98/03/17  */

  double  tmp1, tmp2, tmp3, tmp4;
	double temp_test;


	{
	  register struct s_typetrans *Trans = &(global->mostrans[itrans]);
	  register struct s_typegeo *Geo = &(global->mosgeo[itrans]);
	  register struct s_typemodel *Model = &(global->mosmodels[Trans->imodel]);


/*  NOTE:  For the intrinsic model of the current and capacitances, 'electrical'
		   voltages VG, VD, VS, referred to the bulk, need to be computed from
		   the 'topological' voltages. At the end intrinsic current and
		   capacitances need to be reconverted to their 'topological' meaning.
		   Diode currents and capacitances can be directly computed in
		   'topological' terms.
*/

	  if (Model->mostype == 'N')   /*  NMOS transistor  */
      {
		v1 = vn1 - vnb;
        v2 = vn2 - vnb;
        VG = vng - vnb;
      }
      else                         /*  PMOS transistor  */
      {
        v1 = vnb - vn1;
        v2 = vnb - vn2;
		VG = vnb - vng;
      }

	  if (v2 >= v1)   /*  'Normal' bias, n2 is the drain, n1 the source    */
	  {
		VS = v1;
        VD = v2;
      } else         /*  'Inverted' bias, n1 is the drain, n2 the source  */
      {
        VS = v2;
		VD = v1;
      }


/****************************************/
/*     STATIC MODEL EQUATIONS           */
/****************************************/

/*  Short-distance matching   !!  mb  97/07/18  */

	  VTOa = VTO + AVTO * AWL;
      GAMMAa = GAMMA + AGAMMA * AWL;
	  KPa = KP * (1.0 + AKP * AWL);

	  if (GAMMAa < 0.0)  GAMMAa = 0.0;        /*  !!  mb  97/09/15  introduced clipping */
      if (KPa < 0.0)  KPa = 0.0;

      GAMMA_sqrt_PHI = GAMMAa*sqrt_PHI;


/*  Pinch-off voltage, including short and narrow channel effects:  */

      /*  VGprime:  */
      /*  !!  mb  97/04/14  introduced RSCE equations (v2.5)  */
	  if (V0 == 0.0)
        deltaVFB = 0.0;
	  else
      {
        /*  !!  mb  99/03/26  corrected for multiple device number  */
        vL = c_a * (Leff/(LK*NS) - 0.1);
        sqv = 1.0 / (1.0 + 0.5*(vL + sqrt(vL*vL + c_ee)));
        deltaVFB = V0 * sqv * sqv;
      }


      /*  NOTE: New VGprime formulation (REVISION III) allows VP to be
                expressed with a single equation
      */

      /*  !!  mb  99/05/10  (r12) */
      VGstar = VG - VTOa - deltaVFB + PHI + GAMMA_sqrt_PHI;
      sqrt_VGstar = sqrt(VGstar*VGstar + 2.0*Vt_Vt_16);             /*  !!  mb  99/09/09  */

      VGprime = 0.5*(VGstar + sqrt_VGstar);


	  /*  Pinch-off voltage VP, limited to VP >= -PHI  */

      PHI_VD = PHI+VD;
      sqrt_PHI_VD_Vt = sqrt(PHI_VD*PHI_VD+Vt_Vt_16);
      sqrt_PHI_VD = sqrt(0.5*(PHI_VD+sqrt_PHI_VD_Vt));

	  PHI_VS = PHI+VS;
      sqrt_PHI_VS_Vt = sqrt(PHI_VS*PHI_VS+Vt_Vt_16);
      sqrt_PHI_VS = sqrt(0.5*(PHI_VS+sqrt_PHI_VS_Vt));

      /*  !!  mb  97/09/14  symmetric version of GAMMAprime necessary with charges model */
      big_sqrt_VP0 = sqrt(VGprime + 0.25*GAMMAa*GAMMAa);
      VP0 = VGprime - PHI - GAMMAa*(big_sqrt_VP0 - 0.5*GAMMAa);
      sqrt_PHI_VP0 = sqrt(VP0+PHI+Vt_01);
	  GAMMAstar = GAMMAa - LETA_L * (sqrt_PHI_VS+sqrt_PHI_VD) + WETA_W * sqrt_PHI_VP0;

	  /*  keep GAMMAprime from becoming negative  */
      sqrt_GAMMAstar = sqrt(GAMMAstar*GAMMAstar+Vt_01);

      GAMMAprime = 0.5*(GAMMAstar+sqrt_GAMMAstar);

	  big_sqrt_VP = sqrt(VGprime+0.25*GAMMAprime*GAMMAprime);

	  VP = VGprime-PHI-GAMMAprime*(big_sqrt_VP-0.5*GAMMAprime);
	  temp_test = VP;
ratioV_if = (temp_test-VS)*inv_Vt;

/*  Forward normalized current:  */

	  ratioV_if = (VP-VS)*inv_Vt;                    /*  !!  mb  95/08/14  */

      /*  !!  mb  96/08/23  interpolation function option  */
	  /*  !!  mb  97/02/10  new interpolation function     */
      if (EKVINT == 0.0)
        F_v_n(ratioV_if, &if_, &sqrt_if, &dif_dv);
      else
        F_v_simple(ratioV_if, &if_, &sqrt_if, &dif_dv);


/*  Saturation voltage:  */

      Vt_Vc = Vt / Vc;
      VDSS_sqrt = sqrt(0.25+sqrt_if*Vt_Vc);

	  VDSS = Vc*(VDSS_sqrt-0.5);

      Vds = 0.5*(VD-VS);
      deltaV_2 = Vt_Vt_16*(LAMBDA*(sqrt_if-VDSS*inv_Vt)+15.625e-3);   /*  !!  mb  95/08/14  */

      sqrt_VDSS_deltaV = sqrt(VDSS*VDSS+deltaV_2);
      sqrt_Vds_VDSS_deltaV = sqrt((Vds-VDSS)*(Vds-VDSS)+deltaV_2);

	  Vip = sqrt_VDSS_deltaV-sqrt_Vds_VDSS_deltaV;

      VDSSprime_sqrt = sqrt(0.25+(sqrt_if-0.75*log(if_))*Vt_Vc);

      VDSSprime = Vc*(VDSSprime_sqrt-0.5)+log_Vc_Vt;


/*  Reverse normalized current:  */

      Vdsprime = Vds-VDSSprime;   /*  !!  mb  97/07/18  introduced Vdsprime  */

      sqrt_VDSSprime_deltaV = sqrt(VDSSprime*VDSSprime+deltaV_2);
      sqrt_Vds_VDSSprime_deltaV = sqrt(Vdsprime*Vdsprime+deltaV_2);

	  /*  !!  mb  95/08/14,  95/08/14  */
      ratioV_irprime = (VP-Vds-VS-sqrt_VDSSprime_deltaV+sqrt_Vds_VDSSprime_deltaV)*inv_Vt;

      /*  !!  mb  96/07/18  interpolation function option  */
      /*  !!  mb  97/02/10  new interpolation function     */
      if (EKVINT == 0.0)
		F_v_n(ratioV_irprime, &irprime, &sqrt_irprime, &dirprime_dv);
      else
        F_v_simple(ratioV_irprime, &irprime, &sqrt_irprime, &dirprime_dv);


/*  Channel length modulation & mobility reduction due to longitudinal field:  */

	  deltaL = Lc_LAMBDA*log(1.0+(Vds-Vip)/Lc_UCRIT);

      Lprime = Leff-deltaL+(Vds+Vip)*inv_UCRIT;    /*  !!  mb  96/04/12  (gg) Leff instead of L  */
	  Lmin = 0.1*Leff;                             /*  !!  mb  96/04/12  (gg) Leff instead of L  */


      /*  !!  mb  96/04/12  correction for sqrt_Lprime_Lmin and Leq  */

      sqrt_Lprime_Lmin = sqrt(Lprime*Lprime+Lmin*Lmin);

      Leq = 0.5*(Lprime+sqrt_Lprime_Lmin);


/*  Transconductance factor:  */

      /*  Mobility reduction due to vertical field  */

      /*  Reverse normalized current: */
      ratioV_ir = (VP-VD) * inv_Vt;
      if (EKVINT == 0.0)
		F_v_n(ratioV_ir, &ir, &sqrt_ir, &dir_dv);
	  else
        F_v_simple(ratioV_ir, &ir, &sqrt_ir, &dir_dv);

	  sif2 = 0.25+if_;
	  sir2 = 0.25+ir;
	  sif = sqrt(sif2);
      sir = sqrt(sir2);
      sif_sir_2 = (sif+sir)*(sif+sir);

      VP_PHI_eps = VP+PHI+1.0e-6;
	  sqrt_PHI_VP_2 = 2.0*sqrt(VP_PHI_eps);

      /*  !!  mb  97/08/21  replaced GAMMA by GAMMAa in n_1, n_1_n, QB (sa)  */
      n_1 = GAMMAa/sqrt_PHI_VP_2;
      n_1_n = GAMMAa/(sqrt_PHI_VP_2+GAMMAa);

      /*  Normalized inversion charge  (qi=QI/WLCox)  */
      qi = -(1.0+n_1)*Vt*((two3rds+two3rds)*(sir2+sir*sif+sif2)/(sif+sir) - 1.0);

      /*  Normalized depletion charge  (qb=QB/WLCox), for depletion to inversion  */
	  qb = -0.5*GAMMAa*sqrt_PHI_VP_2 - n_1_n*qi;


      if (E0 == 0.0)
      {

        /*  NOTE: this version of the simple mobility model from prior versions
            of the EKV model is reinstated. In case E0 is *not* specified, this
			simple mobility model is used according to THETA, if specified.
		*/

		/*  VPprime:  */

        /*  !!  mb  eliminated discontinuity of derivative of 1+THETA*VP  */
        sqrt_VP_Vt = sqrt(VP*VP + Vt_Vt_2);
        VPprime = 0.5 * (VP + sqrt_VP_Vt);

        THETA_VP_1 = 1.0+THETA*VPprime;

        beta = KPa * Weff / (Leq * THETA_VP_1);    /*  !!  mb  97/07/18  */

      }
      else
      {
		/*  new model for mobility reduction, linked to the charges model  */

        /*  !!  mb  98/10/11  (r10)  introduced fabs(Eeff) (jpm)  */
        E0_Q_1 = 1.0 + T0*fabs(qb+eta_qi*qi);
        T0_GAMMA_1 = 1.0 + T0*GAMMA_sqrt_PHI;   /*  !!  mb  97/06/02  ekv v2.6  */

        beta = KPa * Weff * T0_GAMMA_1 / (Leq * E0_Q_1);    /*  !!  mb  97/07/18  */

     }


/*  Slope factor:  */

	  /*  !!  mb introduced new formula to avoid divergence of n for VP->-PHI  */

      sqrt_PHI_VP = sqrt(PHI+VP+Vt_4);   /*  !!  mb  95/12/19  introduced Vt_4  */

      n = 1.0 + GAMMAa/(2.0*sqrt_PHI_VP);


/*  Drain current:  */

      if_ir = if_-irprime;
      Ispec = Vt_Vt_2 * n * beta;

      IDS = Ispec * if_ir;


/*******     ASSIGN COMPUTED CURRENT    *******/

      global->current = IDS;

      if (v1 > v2)
    	  global->current = - global->current;

      if (Model->mostype == 'P')          /*  PMOS transistor  */
    	  global->current = - global->current;

/*  NOTE:  Until here the 'electrical' current has been computed and needs to
		   be converted to its 'topological' meaning, i.e. if source and drain
		   voltages have been permuted, the current needs to change sign
           For PMOS, as voltages signs have all been changed on entry, all currents
           change sign.
*/



/*******        START OF ANALYTIC DERIVATIVES COMPUTATION                *******/

/*  NOTE:  Some simulators calculate derivatives only numerically. In this case
           the following part on analytical derivatives should be suppressed. In
           some cases, simulators let the user choose if analytical or numerical
		   derivatives should be used. The following flag needs to be set
           internally by the simulator. 'analytic_deriv' should be set only when
           the analytical derivatives are really needed.
*/

      if (global->analytic_deriv)
      {

/*  !!  mb  97/07/18  changed intermediate variables for derivatives computation (tmp1, ...
                      instead of d...._factor, ...)
*/

/*  Pinch-off voltage derivatives:  */

      /*  !!  mb  97/09/14  symmetric version of GAMMAprime necessary with charges model
          !!  mb  99/05/10  (r12) New VGprime formulation (REVISION III) allows
							VP derivatives to be expressed with a single equation
      */

      tmp1 = GAMMAprime / (sqrt_GAMMAstar+sqrt_GAMMAstar);
      tmp2 = VGprime/sqrt_VGstar;                            /*  dVGprime_dVG  */

      dGAMMAprime_dVD = -LETA_L * tmp1 * sqrt_PHI_VD / sqrt_PHI_VD_Vt;
      dGAMMAprime_dVS = -LETA_L * tmp1 * sqrt_PHI_VS / sqrt_PHI_VS_Vt;
      dGAMMAprime_dVG =  WETA_W * tmp1 * (big_sqrt_VP0-0.5*GAMMAa)
                                                     / (big_sqrt_VP0*sqrt_PHI_VP0) * tmp2;

      tmp3 = (VP+PHI) / big_sqrt_VP;

      dVP_dVD = -tmp3 * dGAMMAprime_dVD;
      dVP_dVS = -tmp3 * dGAMMAprime_dVS;
      dVP_dVG = -tmp3 * dGAMMAprime_dVG + (1.0 - GAMMAprime/(big_sqrt_VP+big_sqrt_VP)) * tmp2;


/*  Forward normalized current derivatives:  */

      tmp1 = dif_dv * inv_Vt;   /*  !!  mb  95/08/28, 97/04/21 */

	  dif_dVD = tmp1 * dVP_dVD;
	  dif_dVS = tmp1 * (dVP_dVS-1.0);
      dif_dVG = tmp1 * dVP_dVG;


/*  Saturation voltage derivatives:  */

      tmp1 = Vt / (4.0*VDSS_sqrt*sqrt_if);

      dVDSS_dVD = tmp1 * dif_dVD;
      dVDSS_dVS = tmp1 * dif_dVS;
      dVDSS_dVG = tmp1 * dif_dVG;


	  /*  deltaV derivatives:  */
      tmp1 = (Vt_4+Vt_4) * LAMBDA;
      tmp2 = Vt / (sqrt_if+sqrt_if);

      ddeltaV_dVD = tmp1 * (dif_dVD*tmp2 - dVDSS_dVD);
      ddeltaV_dVS = tmp1 * (dif_dVS*tmp2 - dVDSS_dVS);
      ddeltaV_dVG = tmp1 * (dif_dVG*tmp2 - dVDSS_dVG);


      /*  Vip derivatives:  */
      tmp1 = 1.0 / sqrt_VDSS_deltaV;
	  tmp2 = 1.0 / sqrt_Vds_VDSS_deltaV;
	  tmp3 = Vds-VDSS;

      dVip_dVD = (VDSS*dVDSS_dVD + ddeltaV_dVD) * tmp1
                - (tmp3 * (0.5-dVDSS_dVD) + ddeltaV_dVD) * tmp2;

      dVip_dVS = (VDSS*dVDSS_dVS + ddeltaV_dVS) * tmp1
				- (tmp3 * (-0.5-dVDSS_dVS) + ddeltaV_dVS) * tmp2;

	  dVip_dVG = (VDSS*dVDSS_dVG + ddeltaV_dVG) * tmp1
                - (tmp3 * -dVDSS_dVG + ddeltaV_dVG) * tmp2;


	  /*  VDSSprime derivatives:  */
      tmp1 = Vt * (sqrt_if-1.5)/(4.0*VDSSprime_sqrt*if_);

      dVDSSprime_dVD = tmp1 * dif_dVD;
      dVDSSprime_dVS = tmp1 * dif_dVS;
      dVDSSprime_dVG = tmp1 * dif_dVG;


/*  Reverse normalized current derivatives:  */

      tmp1 = dirprime_dv * inv_Vt;   /*  !!  mb  95/08/28,  97/04/21  */
      tmp2 = 1.0 / sqrt_VDSSprime_deltaV;            /*  !!  mb  97/04/21  */
      tmp3 = 1.0 / sqrt_Vds_VDSSprime_deltaV;

	  dirprime_dVD = tmp1 * (dVP_dVD-0.5
		  -(VDSSprime*dVDSSprime_dVD+ddeltaV_dVD) * tmp2
          +(Vdsprime*(0.5-dVDSSprime_dVD)+ddeltaV_dVD) * tmp3);

      dirprime_dVS = tmp1 * (dVP_dVS-0.5
          -(VDSSprime*dVDSSprime_dVS+ddeltaV_dVS) * tmp2
		  +(Vdsprime*(-0.5-dVDSSprime_dVS)+ddeltaV_dVS) * tmp3);

      dirprime_dVG = tmp1*(dVP_dVG
          -(VDSSprime*dVDSSprime_dVG+ddeltaV_dVG) * tmp2
		  +(Vdsprime*(-dVDSSprime_dVG)+ddeltaV_dVG) * tmp3);


/*  Channel length modulation & mobility reduction derivatives:  */

      /*  deltaL derivatives:   */
      tmp1 = Lc_LAMBDA / (Lc_UCRIT+Vds-Vip);

      ddeltaL_dVD = tmp1 * (0.5-dVip_dVD);
      ddeltaL_dVS = tmp1 * (-0.5-dVip_dVS);
      ddeltaL_dVG = -tmp1 * dVip_dVG;


      /*  Leq derivatives:  */
	  /*  !!  mb   97/09/14 corrected derivatives (sa)  */
	  tmp1 = 1.0 / sqrt_Lprime_Lmin;

	  /*  in fact dLeq_dVX/Leq  */
      dLeq_dVD = tmp1 * (-ddeltaL_dVD + (0.5+dVip_dVD)*inv_UCRIT);
      dLeq_dVS = tmp1 * (-ddeltaL_dVS + (-0.5+dVip_dVS)*inv_UCRIT);
      dLeq_dVG = tmp1 * (-ddeltaL_dVG + dVip_dVG*inv_UCRIT);


/*  Transconductance factor derivatives:  */

      tmp1 = dir_dv*inv_Vt;

      dir_dVD = tmp1 * (dVP_dVD-1.0);
	  dir_dVS = tmp1 * dVP_dVS;
      dir_dVG = tmp1 * dVP_dVG;

      tmp1 = -(1.0+n_1)*Vt*two3rds/sif_sir_2;
      tmp2 = tmp1*(sif+2.0*sir);
      tmp3 = tmp1*(sir+2.0*sif);

      /*  !!  mb  97/09/04  replaced QI by qi  (dQI_dVx are normalized to WLCOX)  */
      tmp1 = -n_1*qi/((2.0+n_1+n_1)*VP_PHI_eps);

      dQI_dVD = tmp1 * dVP_dVD + tmp2 * dif_dVD + tmp3 * dir_dVD;
      dQI_dVS = tmp1 * dVP_dVS + tmp2 * dif_dVS + tmp3 * dir_dVS;
      dQI_dVG = tmp1 * dVP_dVG + tmp2 * dif_dVG + tmp3 * dir_dVG;

	  /*  !!  mb  97/09/04  replaced QI by qi  (dQB_dVx are normalized to WLCOX)  */
	  tmp1 = (1.0+n_1)-qi/(2.0*(1.0+n_1)*VP_PHI_eps);

      dQB_dVD = -n_1_n * (tmp1 * dVP_dVD + dQI_dVD);
      dQB_dVS = -n_1_n * (tmp1 * dVP_dVS + dQI_dVS);
      dQB_dVG = -n_1_n * (tmp1 * dVP_dVG + dQI_dVG);

      if (E0 == 0.0)
      {

		tmp1 = THETA * VPprime / (THETA_VP_1 * sqrt_VP_Vt);

        /*  VPprime derivatives:  */
        dVPprime_dVD = tmp1 * dVP_dVD;
		dVPprime_dVS = tmp1 * dVP_dVS;
        dVPprime_dVG = tmp1 * dVP_dVG;

        dbeta_dVD = -dLeq_dVD - dVPprime_dVD;      /*  in fact dbeta_dVX / beta  */
        dbeta_dVS = -dLeq_dVS - dVPprime_dVS;
        dbeta_dVG = -dLeq_dVG - dVPprime_dVG;

      }
      else
      {

		tmp1 = T0 / E0_Q_1;

		/*  in fact dbeta_dVX / beta  */
        /*  !!  mb   97/08/21 corrected derivatives (sa)  */
        dbeta_dVD = -dLeq_dVD + tmp1 * (dQB_dVD+eta_qi*dQI_dVD);
        dbeta_dVS = -dLeq_dVS + tmp1 * (dQB_dVS+eta_qi*dQI_dVS);
        dbeta_dVG = -dLeq_dVG + tmp1 * (dQB_dVG+eta_qi*dQI_dVG);

      }


/*  Slope factor derivatives:  */

      /*  !!  mb  97/08/21  replaced GAMMA by GAMMAa (sa)  */
      tmp1 = -GAMMAa/(4.0*n*sqrt_PHI_VP*(PHI+VP+Vt_4));    /*  !!  mb  95/12/19  */

      dn_dVD = tmp1 * dVP_dVD;
	  dn_dVS = tmp1 * dVP_dVS;
      dn_dVG = tmp1 * dVP_dVG;


/*  Transconductances:  */

	  gds =  Ispec*((dn_dVD + dbeta_dVD)*if_ir + dif_dVD - dirprime_dVD);

      gms = -Ispec*((dn_dVS + dbeta_dVS)*if_ir + dif_dVS - dirprime_dVS);

	  gm =   Ispec*((dn_dVG + dbeta_dVG)*if_ir + dif_dVG - dirprime_dVG);

      gmbs = gms - gm - gds;



/*******      ASSIGN DERIVATIVES    *******/

      global->dids_dvg1 = gm;                /*  di21_dvg1 = di21_dvgb  */
      global->dids_dvb1 = gmbs;
      global->dids_dv21 = gds;               /*  di21_dv21 = di21_dv2b  */
                                     /*  di21_dvb1 = -(di21_dv2b+di21_dv1b+di21_dvgb) */
      if (v1 > v2)
      {
    	  global->dids_dvg1 = -global->dids_dvg1;
    	  global->dids_dvb1 = -global->dids_dvb1;
    	  global->dids_dv21 = gms;             /*  gms = gds+gm+gmbs  */
      }

/*  NOTE:  The fourth derivative of the current is the negative of the sum of the
           three other derivatives, which would result in -(gm+gmbs+gds) = -gms.

           Until here the 'electrical' variables have been computed and need to
           be converted to 'topological' ones.

		   In case source and drain voltages have been permuted, the conductances
		   dids_dvg1 and dids_dvb1 change sign, dids_dv21 and the fourth derivative
		   are permuted and change sign
*/
      }  /*  end  if (analytic_derivatives)  ...  */


/****  END OF ANALYTIC DERIVATIVES COMPUTATION  ****/


/******  Store gm and noise conductance for noise analysis  ******/

/*  !!  mb  98/07/09  changed function call  */

      Geo->gm = gm;                    /*  used for flicker noise  */
	  Geo->gn = beta * fabs(qi);       /*  used for thermal noise  */



/*  Impact ionization current:  */
/*  !!  mb  95/12/19  introduced impact ionization
					  This current component is flowing from the intrinsic drain terminal
                      to the bulk (for NMOS) in parallel with the junction current.
                      The simulator should also take into account the corresponding
                      conductances.
*/


	  global->dib1_dv2b = global->dib1_dv1b = global->dib1_dvgb = 0.0;
	  global->dib2_dv2b = global->dib2_dv1b = global->dib2_dvgb = 0.0;

      /*  Substrate current:  */
      Vib = VD-VS-IBN_2*VDSS;   /*  !!  mb  96/07/15  corrected VDSS factor  */

      if ((Vib > 0.0) && (IBA_IBB > 0.0))    /*  !!  mb  98/05/14  */
      {
        /*  !!  mb  96/06/20  intr. Lc_IBB_Vib  */
        inv_Vib = 1.0/Vib;
        Lc_IBB_Vib = -Lc_IBB*inv_Vib;
        exp_ib = exp(Lc_IBB_Vib);
        isub = IBA_IBB*Vib*exp_ib;
		Isub = isub*IDS;
        dIsub_factor = Isub*inv_Vib*(1.0-Lc_IBB_Vib);    /*  !!  mb  97/04/21  corrected sign  */

        if (v2 >= v1)
        	global->ib2 -= Isub;
        else
        	global->ib1 -= Isub;

		if (global->analytic_deriv)
        {

		  /*  !!  mb  97/04/21  completed derivatives w.r.t. VS, VG  */
		  if (v2 >= v1)
		  {

			  global->dib2_dv2b = -isub*gds - dIsub_factor*(1.0-IBN_2*dVDSS_dVD);
			  global->dib2_dv1b =  isub*gms - dIsub_factor*(-1.0-IBN_2*dVDSS_dVS);
			  global->dib2_dvgb = -isub*gm  - dIsub_factor*(-IBN_2*dVDSS_dVG);

            /*  !!  mb  97/04/21  if the source terminal (1) is taken as reference for the substrate
                                        current derivatives, the following relations hold:
                          dib2_dvg1 = dib2_dvgb;
                          dib2_dv21 = dib2_dv2b;
                          dib2_dvb1 = -(dib2_dv2b+dib2_dv1b+dib2_dvgb);
            */
          }
		  else
          {

			  global->dib1_dv2b = -isub*gds - dIsub_factor*(1.0-IBN_2*dVDSS_dVD);
			  global->dib1_dv1b =  isub*gms - dIsub_factor*(-1.0-IBN_2*dVDSS_dVS);
			  global->dib1_dvgb = -isub*gm  - dIsub_factor*(-IBN_2*dVDSS_dVG);

          }
        }

	  }   /*  end  if ((Vib > 0.0) && (IBA_IBB > 0.0))   */


	  if (Model->mostype == 'P')          /*  PMOS transistor  */
      {
		  global->ib2 = -global->ib2;
		  global->ib1 = -global->ib1;
      }

/*      END OF SUBSTRATE CURRENT COMPUTATION        */


/****  END OF STATIQUE INTRINSIC EQUATIONS  ****/


/************************************************/
/*        QUASISTATIC MODEL EQUATIONS           */
/************************************************/

/*  The dynamic model associated to the EKV model.

    NOTE:  two dynamic models can be chosen depending on the setup parameter XQC
           (choice among charges or capacitances models).

             XQC = 0.4 (default)     complete charges/(trans-)capacitances model
             XQC = 1.0            capacitances model for the 5 intrinsic capacitances

		   There is no other 'charge partitioning' available among drain and source charges
		   than the one implemented, derived from the 'natural' linear integration
		   of the inversion charge densities, leading to 60/40 partinioning in saturation
           among source and drain charges. Note that at VDS=0, symmetrical behavior is
           obtained for source and drain sides.

           If the parameter EKVDYN=0 is set, then all intrinsic capacitances are set
		   to zero, i.e. only the static model is accounted for.

    NOTE:  the following flags need to be set internally by the simulator.
*/


      if (global->transient_analysis || global->ac_analysis)
      {


if (EKVDYN == 0.0)
{

/******  Switch off all 5 intrinsic capacitances  ******/

        Cgsi = Cgdi = Cgbi = Csbi = Cdbi = TAU = 0.0;

}
else
{
/*----------------  NEW DYNAMIC MODEL  -------------------*/


        if (XQC == 1.0)
        {

		  /*  Only the 5 intrinsic capacitances are computed. Capacitances
              are derived from the new charge expressions   !!  mb  96/08/23
              These expressions do not account for the n dependence on VG, VS, VD).
              These intrinsic capacitances should only be computed in case the
              simulator demands capacitances rather than charges.
          */

          cgsi = two3rds*(1.0-(sir2+sir+0.5*sif)/sif_sir_2);
		  cgdi = two3rds*(1.0-(sif2+sif+0.5*sir)/sif_sir_2);

          /*  !!  mb  98/08/08  erroneous n_1 term replaced by n_1_n  (jpm)  */
          /*  cgbi = n_1*(1.0-cgsi-cgdi);  */
          cgbi = n_1_n*(1.0-cgsi-cgdi);

          Cgsi = WLCox * cgsi;
          Cgdi = WLCox * cgdi;
          Cgbi = WLCox * cgbi;

          Csbi = Cgsi * n_1;
		  Cdbi = Cgdi * n_1;

		}
        else
        {

          /*  Complete charges/(trans-)capacitances model. The four node charges and the
			  (non-reciprocal) transcapacitances are computed. The transcapacitances are
              the analytical derivatives of the node charges with respect to the bias
              voltages VG, VD, VS.

              NOTE:  there is no other 'charge partitioning' available among drain and source
              charges than the one implemented, derived from the natural linear integration
              of the inversion charge densities, leading to 60/40 partinioning
              in saturation among source and drain charges. Note that all charges and
			  transcapacitances are continuous and symmetrical for if, ir (i.e. for VD, VS).
          */


          sif3 = sif*sif2;
          sir3 = sir*sir2;

          /*  new charges expressions (v2.6 Revision III)  */

              tmp1 = sqrt(PHI+0.5*VP);
              sqrt_PHI_VP2_2 = tmp1+tmp1;
			  n_Vt_COX = (1.0 + GAMMAprime/sqrt_PHI_VP2_2) * Vt*WLCox;

			  QD = -n_Vt_COX*(four15th*(3.0*sir3+6.0*sir2*sif+4.0*sir*sif2+2.0*sif3)/sif_sir_2 - 0.5);
              QS = -n_Vt_COX*(four15th*(3.0*sif3+6.0*sif2*sir+4.0*sif*sir2+2.0*sir3)/sif_sir_2 - 0.5);

              QI = QS+QD;

			  QB = WLCox * (-0.5*GAMMAprime*sqrt_PHI_VP_2 + VGprime - VGstar)
                                                          - QI*GAMMAprime/(GAMMAprime+sqrt_PHI_VP2_2);

          /*  (end) new charges expressions (v2.6 Revision III)  */


          /*  QG = -QI-Qox-QB;  */
          QG = -QI -QB;


		  if (global->analytic_deriv)
          {

            /*  Analytic derivatives of the node charges (transcapacitances)  !!  mb  96/11/28  */

            /*  !!  mb  00/09/15  (gr)  reinstated sif_sir_3 and dQ_i_factor   */

            sif_sir_3 = (sif+sir)*sif_sir_2;
            dQ_i_factor = -n_Vt_COX*four15th/sif_sir_3;


			/*  new charges derivatives (v2.6 Revision III)  */

              tmp1 = sir2+3.0*sir*sif+sif2;
              tmp2 = 1.5*tmp1+2.5*sif2;
              tmp3 = QD/(GAMMAprime+sqrt_PHI_VP2_2);
			  tmp4 = 0.5*GAMMAprime/(PHI+PHI+VP);

			  dQD_dVD = tmp3*(dGAMMAprime_dVD-tmp4*dVP_dVD) + dQ_i_factor*(tmp1*dif_dVD + tmp2*dir_dVD);
              dQD_dVS = tmp3*(dGAMMAprime_dVS-tmp4*dVP_dVS) + dQ_i_factor*(tmp1*dif_dVS + tmp2*dir_dVS);
              dQD_dVG = tmp3*(dGAMMAprime_dVG-tmp4*dVP_dVG) + dQ_i_factor*(tmp1*dif_dVG + tmp2*dir_dVG);

              tmp2 = 1.5*tmp1+2.5*sir2;
              tmp3 = QS/(GAMMAprime+sqrt_PHI_VP2_2);

              dQS_dVD = tmp3*(dGAMMAprime_dVD-tmp4*dVP_dVD) + dQ_i_factor*(tmp2*dif_dVD + tmp1*dir_dVD);
              dQS_dVS = tmp3*(dGAMMAprime_dVS-tmp4*dVP_dVS) + dQ_i_factor*(tmp2*dif_dVS + tmp1*dir_dVS);
              dQS_dVG = tmp3*(dGAMMAprime_dVG-tmp4*dVP_dVG) + dQ_i_factor*(tmp2*dif_dVG + tmp1*dir_dVG);


              dQI_dVD = dQD_dVD+dQS_dVD;
              dQI_dVS = dQD_dVS+dQS_dVS;
              dQI_dVG = dQD_dVG+dQS_dVG;


			  tmp1 = GAMMAprime + sqrt_PHI_VP2_2;
			  tmp2 = sqrt_PHI_VP2_2/(tmp1*tmp1);
			  tmp3 = GAMMAprime/tmp1;
			  tmp4 = -WLCox*0.5*sqrt_PHI_VP_2 - QI*tmp2;
              tmp2 = GAMMAprime * (QI/(tmp1*tmp1*sqrt_PHI_VP2_2) - WLCox/sqrt_PHI_VP_2);


			  dQB_dVD = tmp4 * dGAMMAprime_dVD + tmp2 * dVP_dVD - tmp3 * dQI_dVD;
              dQB_dVS = tmp4 * dGAMMAprime_dVS + tmp2 * dVP_dVS - tmp3 * dQI_dVS;
              dQB_dVG = tmp4 * dGAMMAprime_dVG + tmp2 * dVP_dVG - tmp3 * dQI_dVG
                                                                + WLCox*(VGprime/sqrt_VGstar - 1.0);

            /*  (end) new charges derivatives (v2.6 Revision III)  */

            /*  old charges derivatives (v2.6 Revision < III)

            dQB_dVD *= WLCox;
            dQB_dVS *= WLCox;
            dQB_dVG = (dQB_dVG + VGprime/sqrt_VGstar - 1.0) * WLCox;


            dQD_n_factor = -0.5*n_1_n*QD/VP_PHI_eps;
            tmp1 = sir2+3.0*sir*sif+sif2;
            tmp2 = 1.5*tmp1+2.5*sif2;

            dQD_dVD = dQD_n_factor*dVP_dVD + dQ_i_factor*(tmp1*dif_dVD + tmp2*dir_dVD);
			dQD_dVS = dQD_n_factor*dVP_dVS + dQ_i_factor*(tmp1*dif_dVS + tmp2*dir_dVS);
			dQD_dVG = dQD_n_factor*dVP_dVG + dQ_i_factor*(tmp1*dif_dVG + tmp2*dir_dVG);


            dQS_n_factor = -0.5*n_1_n*QS/VP_PHI_eps;
            tmp2 = 1.5*tmp1+2.5*sir2;

			dQS_dVD = dQS_n_factor*dVP_dVD + dQ_i_factor*(tmp2*dif_dVD + tmp1*dir_dVD);
            dQS_dVS = dQS_n_factor*dVP_dVS + dQ_i_factor*(tmp2*dif_dVS + tmp1*dir_dVS);
            dQS_dVG = dQS_n_factor*dVP_dVG + dQ_i_factor*(tmp2*dif_dVG + tmp1*dir_dVG);

            .... (end) old charges derivatives (v2.6 Revision < III)  */


            dQG_dVD = -dQD_dVD - dQS_dVD - dQB_dVD;
			dQG_dVS = -dQD_dVS - dQS_dVS - dQB_dVS;
            dQG_dVG = -dQD_dVG - dQS_dVG - dQB_dVG;


/*  Notes on the meaning of the variables and conversion depending on the
    reference of the voltages:

  With the following denominations....

     derivatives of charge Qx vs. vgb, vdb, vsb:        cxgb, cxdb, cxsb
     derivatives of charge Qx vs. vgs, vds, vbs:        cxgs, cxds, cxbs
	 derivatives of charge Qx vs. vgr, vdr, vsr, vbr:   cxg, cxd, cxs, cxb

	 (where X=G,B,D,S, and r is an arbitrary reference point)

  ..... the correspondences can be established:

     cxg = cxgb              = cxgs
	 cxd = cxdb              = cxds
     cxs = cgsb              = -(cxgs+cxbs+cxds)
     cxb = -(cxgb+cxdb+cxsb) = cxbs

  which allow to form the total of 16 transcapacitances.

     dQG_dVG:  cggb
     dQG_dVD:  cgdb
	 dQG_dVS:  cgsb

     dQB_dVG:  cbgb
     dQB_dVD:  cbdb
     dQB_dVS:  cbsb

     dQD_dVG:  cdgb
     dQD_dVD:  cddb
     dQD_dVS:  cdsb

     dQS_dVG:  csgb
	 dQS_dVD:  csdb
	 dQS_dVS:  cssb

*/

          }  /*  end  if (analytic_derivatives)  */

		}  /*  end  if (XQC == 0.0)  else  ....  */

        /*  Intrinsic time constant for non-quasi-static model  */

        tau_zero = WLCox/(beta*Vt_2);     /*  !!  mb  98/07/09  */

        if (!global->analytic_deriv)
          sif_sir_3 = (sif+sir)*sif_sir_2;

        TAU = tau_zero*four15th*(sif_sir_2+sif*sir)/sif_sir_3;



/*--------------  END NEW DYNAMIC MODEL  -----------------*/

}  /*  end  if (EKVDYN == 0.0)  else  ....  */


      }  /*  end  if (transient_analysis || ac_analysis)  */



/*        OPERATING POINT INFORMATION        */

      if (global->operating_point)          /*  Note: set flag for operating point info  */
      {

        printf("IDS    %E\n", IDS);            /*  Drain-source current           */
        printf("GM     %E\n", gm);             /*  Gate transconductance          */
        printf("GMS    %E\n", gms);            /*  Source transconductance        */
        printf("GMBS   %E\n", gmbs);           /*  Source transconductance        */
        printf("GDS    %E\n", gds);            /*  Output conductance             */
		printf("VG     %E\n", VG);             /*  Gate-bulk voltage              */
        printf("VD     %E\n", v1);             /*  Drain-bulk voltage             */
		printf("VS     %E\n", v2);             /*  Source-bulk voltage            */
        printf("VP     %E\n", VP);             /*  Pinch-off voltage              */
		printf("n      %E\n", n);              /*  Slope factor                   */
		printf("nVP_VS %E\n", n*(VP-VS));      /*  n*(VP-VS) ~ VGS-VT0            */

		/*  !!  mb  98/03/17  introduced 'SPICE-like' VTH and VDSAT calculation for
							  operating point information
			!!  mb  98/08/08  corrected VTH to include charge-sharing effect
		*/

		VTH = VTO + deltaVFB + GAMMAprime*sqrt_PHI_VS - GAMMAa*sqrt_PHI;   /*  !!  mb  98/08/08  */

		VDSAT = VDSS + VDSS + Vt_4;                                        /*  !!  mb  98/08/20  */

		if (Model->mostype == 'P')          /*  PMOS transistor  */
		{
		  VTH *= -1.0;
		  VDSAT *= -1.0;
		}

		printf("VTH    %E\n", VTH);            /*  'Spice-like' Vth               */
		printf("VDSS   %E\n", VDSAT);          /*  VDSS expression                */

		printf("2nBVt2 %E\n", Ispec);          /*  Specific current               */
		printf("if     %E\n", if_);            /*  Normalized forward current     */
		printf("ir     %E\n", ir);             /*  Normalized reverse current     */
		if (IDS!=0.0)
		  printf("tef    %E\n", gms*Vt/IDS);   /*  Transcond. efficiency factor   */
		printf("VM     %E\n", IDS/gds);        /*  Early voltage                  */
		printf("tau0   %E\n", tau_zero);       /*  Weak inversion transit time    */
		printf("tau    %E\n", TAU);            /*  Transadmittance time constant  */
		if (if_/ir > SATLIM)
		  printf("SAT\n");                     /*  Saturation                     */
		else
		  printf("NON-SAT\n");                 /*  Conduction                     */

	  }  /*  end if (operating_point)  */

  }   /*  end of body of ekv_moseval  */

  return IDS;
}   /*  end  ekv_moseval  */


// -----------MOJE FUNKCJE -----------------------------------------------

#define VISIBLE_PARAMETERS_NUMBER 10

void setDefaultEKVParametersValues (double * _point, unsigned int size)
{
	// values taken from init_param
	_point[0] = 0.6;			//Model->vto      /*  nominal threshold voltage               */
	_point[1] = 0.71;			//Model->gamma    /*  body effect parameter                   */
	_point[2] = 0.97;			//Model->phi      /*  bulk Fermi potential                    */
	_point[3] = 1.50E-04;		//Model->kp       /*  transconductance parameter              */
	_point[4] = 0.1;			//Model->theta    /*  mobility reduction coefficient (reinstated, optional)  */
	_point[5] = 4.50E+06;		//Model->ucrit    /*  longitudinal critical field             */
	_point[6] = NAN;			//Model->cox      /*  gate oxide capacitance                  */
	_point[7] = NAN;			//Model->tox      /*  gate oxide thickness (optional)         */
	_point[8] = 0.5E-6;			//Trans->l        /*  channel length                     */
	_point[9] = 0.8E-6;			//Trans->w        /*  channel width                      */
}

//---------------------------------
// funkcja pobiera jako parametr tablice argumentow
// zwraca wartosc pradu
double ekv_current_count(const double * ekv_param, unsigned int size, double VG, double VD, double VS, double VB)
{
	double IDS;
	typemodel mosmodels_data;
	typetrans typetrans_data;
	typegeo   typegeo_data;
	model_global_data_t global;

	if (size != VISIBLE_PARAMETERS_NUMBER) {
		fprintf(stderr, "ekv_current_count: wrong number of EKV model parameters\n");
		abort();
	}

	global.mosmodels = &mosmodels_data;
	global.mostrans =  &typetrans_data;
	global.mosgeo =    &typegeo_data;

    // moja funkcja inicjacji parametrow
	init_param(&global, ekv_param, model_index,trans_index);

    // inicjowanie Weff, Leff i def W i L - jesli ich brak
	ekv_initdefaultgeo(&global, trans_index);

    // testowanie poprawnosci parametrow
	ekv_initmodelparams(&global, model_index);
	ekv_initmodel(&global, model_index);
	ekv_initgeo(&global, trans_index);


	// model's flags setup
	global.access_res         = no;             /*  presence of access resistors        */
	global.analytic_deriv     = no;         /*  analytical derivatives computation  */
	global.newton_analysis    = no;        /*  Newton analysis                     */
	global.transient_analysis = no;     /*  transient analysis                  */
	global.ac_analysis        = no;            /*  ac analysis                         */

	global.operating_point    = no;        /*  operating point                     */
	global.inv_polar          = no;              /*  inverse polarisation                */

	// VG, VD, VS, VB)
	return ekv_moseval(&global, trans_index, VG, VD, VS,VB);
}
